import numpy as np
import pyccl as ccl
import fastpt as fpt

cosmo = ccl.Cosmology(Omega_c=0.25, Omega_b=0.05, h=0.7, n_s=0.96, sigma8=0.8,
                      transfer_function='bbks')
lkmin = -4
lkmax = 2
n_per_decade = 20
z_s = np.array([0., 1.])
a_s = 1./(1 + z_s)
ks = np.logspace(lkmin, lkmax, (lkmax - lkmin) * n_per_decade)
pk = ccl.linear_matter_power(cosmo, ks, 1.)
gf = ccl.growth_factor(cosmo, a_s)
g4 = gf**4
pklin = np.array([ccl.linear_matter_power(cosmo, ks, a)
                  for a in a_s])

C_window=.75
P_window=None
n_pad=int(0.5*len(ks))
to_do = ['one_loop_dd', 'dd_bias', 'IA']
pt_ob=fpt.FASTPT(ks,to_do=to_do,low_extrap=-5,high_extrap=3,n_pad=n_pad)
oloop_dd = pt_ob.one_loop_dd(pk,
                             P_window=P_window,
                             C_window=C_window)
Pd1d1 = pklin + g4[:, None] * oloop_dd[0][None, :]
dd_bias = pt_ob.one_loop_dd_bias_b3nl(pk,
                                      P_window=P_window,
                                      C_window=C_window)

ia_ta = pt_ob.IA_ta(pk,
                    P_window=P_window,
                    C_window=C_window)
ia_tt = pt_ob.IA_tt(pk,
                    P_window=P_window,
                    C_window=C_window)
ia_mix = pt_ob.IA_mix(pk,
                      P_window=P_window,
                      C_window=C_window)

g4 = g4[:, None]
Pd1d2 = g4 * dd_bias[2][None, :]
Pd2d2 = g4 * dd_bias[3][None, :]
Pd1s2 = g4 * dd_bias[4][None, :]
Pd2s2 = g4 * dd_bias[5][None, :]
Ps2s2 = g4 * dd_bias[6][None, :]
Pd1d3 = g4 * dd_bias[8][None, :]
Pd1k2 = Pd1d1 * (ks**2)[None, :]
a00e = g4 * ia_ta[0][None, :]
c00e = g4 * ia_ta[1][None, :]
a0e0e = g4 * ia_ta[2][None, :]
a0b0b = g4 * ia_ta[3][None, :]
ae2e2 = g4 * ia_tt[0][None, :]
ab2b2 = g4 * ia_tt[1][None, :]
a0e2 = g4 * ia_mix[0][None, :]
b0e2 = g4 * ia_mix[1][None, :]
d0ee2 = g4 * ia_mix[2][None, :]
d0bb2 = g4 * ia_mix[3][None, :]
                                        
b1 = 1.3
b2 = 1.5
bs = 1.7
b3 = 1.9
bk2 = 0.1
c1 = 1.9
c2 = 2.1
cd = 2.3

pgg = (b1**2 * Pd1d1 +
       b1*b2 * Pd1d2 +
       0.25 * b2**2 * Pd2d2 +
       b1 * bs * Pd1s2 +
       0.5 * b2 * bs * Pd2s2 +
       0.25 * bs**2 * Ps2s2 +
       b1 * b3 * Pd1d3 +
       b1 * bk2 * Pd1k2)

pgm = (b1 * Pd1d1 +
       0.5* b2 * Pd1d2 +
       0.5 * bs * Pd1s2 +
       0.5 * b3 * Pd1d3 +
       0.5 * bk2 * Pd1k2)
pgi = b1 * (c1 * Pd1d1 +
            cd * (a00e + c00e) +
            c2 * (a0e2 + b0e2))
pii = (c1**2 * Pd1d1 +
       2 * c1 * cd * (a00e + c00e) +
       cd**2 * a0e0e +
       c2**2 * ae2e2 +
       2 * c1 * c2 * (a0e2 + b0e2) +
       2 * cd * c2 * d0ee2)
pii_bb = (cd**2 * a0b0b +
          c2**2 * ab2b2 +
          2 * cd * c2 * d0bb2)
pim = (c1 * Pd1d1 +
       cd * (a00e + c00e) +
       c2 * (a0e2 + b0e2))

np.savetxt("../pt_bm_z0.txt",
           np.transpose([ks, pgg[0], pgm[0], pgi[0],
                         pii[0], pii_bb[0], pim[0]]),
           header='[0]-k  [1]-GG [2]-GM [3]-GI [4]-II [5]-II_BB [6]-IM')
np.savetxt("../pt_bm_z1.txt",
           np.transpose([ks, pgg[1], pgm[1], pgi[1],
                         pii[1], pii_bb[1], pim[1]]),
           header='[0]-k  [1]-GG [2]-GM [3]-GI [4]-II [5]-II_BB [6]-IM')



"""Config file for pytest. Allow caching for faster test completion.
`$ pytest tests/ --use-cache`
"""


def pytest_addoption(parser):
    parser.addoption("--use-cache", action="store_true", help="Enable cache.")


def pytest_generate_tests(metafunc):
    if metafunc.config.getoption("use_cache"):
        import pyccl
        pyccl.Caching.enable()



import numpy as np
import pytest
import pyccl as ccl

COSMO = ccl.Cosmology(
    Omega_c=0.27, Omega_b=0.045, h=0.67, sigma8=0.8, n_s=0.96,
    transfer_function='bbks', matter_power_spectrum='linear')

COSMO_NU = ccl.Cosmology(
    Omega_c=0.27, Omega_b=0.045, h=0.67, sigma8=0.8, n_s=0.96,
    transfer_function='bbks', matter_power_spectrum='linear', m_nu=0.1,
    Omega_k=0.1)

AVALS = [
    1,
    1.0,
    0.8,
    [0.2, 0.4],
    np.array([0.2, 0.4])]

input_a_array_not1 = np.linspace(0.1, 0.9, 100)
input_a_array_descending = np.linspace(1.0, 0.1, 100)
input_a_array = np.linspace(0.1, 1, 100)
input_chi = ccl.background.comoving_radial_distance(COSMO, input_a_array)
input_hoh0 = ccl.background.h_over_h0(COSMO, input_a_array)
input_growth = ccl.background.growth_factor_unnorm(COSMO, input_a_array)
input_fgrowth = ccl.background.growth_rate(COSMO, input_a_array)


def test_angular_distance_arrs():
    a_lens = 1.0
    a_source = 0.5

    # Pass no lens, one source
    d0 = ccl.angular_diameter_distance(COSMO, a_source)

    # Pass no lens, many sources
    d = ccl.angular_diameter_distance(COSMO, np.full(10, a_source))
    assert np.all(np.fabs(d-d0) < 1E-10)

    # One source, one lens
    d = ccl.angular_diameter_distance(COSMO, a_lens, a_source)
    assert np.fabs(d0-d) < 1E-10

    # Many sources, one lens
    d = ccl.angular_diameter_distance(COSMO, a_lens, np.full(10, a_source))
    assert np.all(np.fabs(d-d0) < 1E-10)

    # Many sources, many lenses
    d = ccl.angular_diameter_distance(COSMO, np.full(10, a_lens),
                                      np.full(10, a_source))
    assert np.all(np.fabs(d-d0) < 1E-10)


@pytest.mark.parametrize('a', AVALS)
@pytest.mark.parametrize('func', [
    ccl.growth_factor,
    ccl.growth_rate,
    ccl.growth_factor_unnorm,
    ccl.scale_factor_of_chi,
    ccl.comoving_angular_distance,
    ccl.comoving_radial_distance,
    ccl.angular_diameter_distance,
    ccl.luminosity_distance,
    ccl.h_over_h0,
    ccl.distance_modulus])
def test_background_a_interface(a, func):
    if func is ccl.distance_modulus and np.any(a == 1):
        with pytest.raises(ccl.CCLError):
            func(COSMO, a)
    else:
        val = func(COSMO, a)
        assert np.all(np.isfinite(val))
        assert np.shape(val) == np.shape(a)
        if (func is ccl.angular_diameter_distance):
            val = func(COSMO, a, a)
            assert np.all(np.isfinite(val))
            assert np.shape(val) == np.shape(a)
            if isinstance(a, (int, float)):
                val1 = ccl.angular_diameter_distance(COSMO, 1., a)
                val2 = ccl.comoving_angular_distance(COSMO, a)*a
            else:
                val1 = ccl.angular_diameter_distance(COSMO, np.ones(len(a)), a)
                val2 = ccl.comoving_angular_distance(COSMO, a)*a
            assert np.allclose(val1, val2)


@pytest.mark.parametrize('a', AVALS)
@pytest.mark.parametrize('kind', [
    'matter',
    'dark_energy',
    'radiation',
    'curvature',
    'neutrinos_rel',
    'neutrinos_massive'])
def test_background_omega_x(a, kind):
    val = ccl.omega_x(COSMO_NU, a, kind)
    assert np.all(np.isfinite(val))
    assert np.shape(val) == np.shape(a)

    if np.all(a == 1):
        if kind == 'matter':
            val_z0 = (
                COSMO_NU['Omega_b'] +
                COSMO_NU['Omega_c'] +
                COSMO_NU['Omega_nu_mass'])
        elif kind == 'dark_energy':
            val_z0 = COSMO_NU['Omega_l']
        elif kind == 'radiation':
            val_z0 = COSMO_NU['Omega_g']
        elif kind == 'curvature':
            val_z0 = COSMO_NU['Omega_k']
        elif kind == 'neutrinos_rel':
            val_z0 = COSMO_NU['Omega_nu_rel']
        elif kind == 'neutrinos_massive':
            val_z0 = COSMO_NU['Omega_nu_mass']

        assert np.allclose(val, val_z0)


def test_background_omega_x_raises():
    with pytest.raises(ValueError):
        ccl.omega_x(COSMO, 1, 'blah')


@pytest.mark.parametrize('a', AVALS)
@pytest.mark.parametrize('kind', [
    'matter',
    'dark_energy',
    'radiation',
    'curvature',
    'neutrinos_rel',
    'neutrinos_massive'])
@pytest.mark.parametrize('is_comoving', [True, False])
def test_background_rho_x(a, kind, is_comoving):
    val = ccl.rho_x(COSMO_NU, a, kind, is_comoving=is_comoving)
    assert np.all(np.isfinite(val))
    assert np.shape(val) == np.shape(a)


def test_background_rho_x_raises():
    with pytest.raises(ValueError):
        ccl.rho_x(COSMO, 1, 'blah', is_comoving=False)


def test_input_arrays():
    cosmo = ccl.Cosmology(Omega_c=0.27, Omega_b=0.05, h=0.7, n_s=0.965,
                          A_s=2e-9)
    # Where the input quantities are calculated.
    a_arr = np.linspace(0.1, 1, 100)
    chi_from_ccl = ccl.background.comoving_radial_distance(cosmo, a_arr)
    hoh0_from_ccl = ccl.background.h_over_h0(cosmo, a_arr)
    growthu_from_ccl = ccl.background.growth_factor_unnorm(cosmo, a_arr)
    fgrowth_from_ccl = ccl.background.growth_rate(cosmo, a_arr)

    background = {'a': a_arr,
                  'chi': chi_from_ccl,
                  'h_over_h0': hoh0_from_ccl}
    growth = {'a': a_arr,
              'growth_factor': growthu_from_ccl,
              'growth_rate': fgrowth_from_ccl}
    cosmo_input = ccl.CosmologyCalculator(Omega_c=0.27, Omega_b=0.05, h=0.7,
                                          n_s=0.965, A_s=2e-9,
                                          background=background,
                                          growth=growth)

    # Where to compare chi(a) from CCL and from CCL with input quantities.
    a_arr = np.linspace(0.102, 0.987, 158)
    chi_ccl_input = ccl.background.comoving_radial_distance(cosmo_input,
                                                            a_arr)
    chi_from_ccl = ccl.background.comoving_radial_distance(cosmo, a_arr)
    # Relative difference (a-b)/b < 1e-5 = 0.001 %
    assert np.allclose(chi_ccl_input, chi_from_ccl, atol=0., rtol=1e-5)

    growth_ccl_input = ccl.background.growth_factor(cosmo_input, a_arr)
    growth_from_ccl = ccl.background.growth_factor(cosmo, a_arr)
    assert np.allclose(growth_ccl_input, growth_from_ccl, atol=0., rtol=1e-5)

    growthu_ccl_input = ccl.background.growth_factor_unnorm(cosmo_input, a_arr)
    growthu_from_ccl = ccl.background.growth_factor_unnorm(cosmo, a_arr)
    assert np.allclose(growthu_ccl_input, growthu_from_ccl, atol=0., rtol=1e-5)

    fgrowth_ccl_input = ccl.background.growth_rate(cosmo_input, a_arr)
    fgrowth_from_ccl = ccl.background.growth_rate(cosmo, a_arr)
    assert np.allclose(fgrowth_ccl_input, fgrowth_from_ccl, atol=0., rtol=1e-5)

    # Test that the distance/growth flags have been set
    assert cosmo_input.has_distances
    assert cosmo_input.has_growth


def test_input_arrays_raises():
    """
    Test for input scale factor array being descending,
    not ending in 1.0, being different size, as well as
    for no input arrays.
    """
    for input_a in [input_a_array_descending,
                    input_a_array_not1,
                    input_a_array[:-2]]:
        with pytest.raises(ValueError):
            ccl.CosmologyCalculator(Omega_c=0.27, Omega_b=0.05, h=0.7,
                                    n_s=0.965, A_s=2e-9,
                                    growth={'a': input_a,
                                            'growth_factor': input_growth,
                                            'growth_rate': input_fgrowth})
        with pytest.raises(ValueError):
            ccl.CosmologyCalculator(Omega_c=0.27, Omega_b=0.05, h=0.7,
                                    n_s=0.965, A_s=2e-9,
                                    background={'a': input_a,
                                                'chi': input_chi,
                                                'h_over_h0': input_hoh0})
    # Not a dictionary
    with pytest.raises(TypeError):
        ccl.CosmologyCalculator(Omega_c=0.27, Omega_b=0.05, h=0.7,
                                n_s=0.965, A_s=2e-9,
                                growth=3)
    with pytest.raises(TypeError):
        ccl.CosmologyCalculator(Omega_c=0.27, Omega_b=0.05, h=0.7,
                                n_s=0.965, A_s=2e-9,
                                background=3)
    # Incomplete dictionary
    with pytest.raises(KeyError):
        ccl.CosmologyCalculator(Omega_c=0.27, Omega_b=0.05, h=0.7,
                                n_s=0.965, A_s=2e-9,
                                background={'a': input_a_array,
                                            'h_over_h0': input_hoh0})
    with pytest.raises(KeyError):
        ccl.CosmologyCalculator(Omega_c=0.27, Omega_b=0.05, h=0.7,
                                n_s=0.965, A_s=2e-9,
                                growth={'a': input_a_array,
                                        'growth_rate': input_fgrowth})



import numpy as np
import pyccl as ccl
import scipy.integrate


def test_hmcalculator_number_counts_smoke():
    cosmo = ccl.Cosmology(
        Omega_c=0.27, Omega_b=0.045, h=0.67, sigma8=0.8, n_s=0.96,
        transfer_function='bbks', matter_power_spectrum='linear')
    mdef = ccl.halos.MassDef(200, 'matter')
    hmf = ccl.halos.MassFuncTinker10(mass_def=mdef, mass_def_strict=False)
    hbf = ccl.halos.HaloBiasTinker10(mass_def=mdef, mass_def_strict=False)

    hmc = ccl.halos.HMCalculator(mass_function=hmf, halo_bias=hbf,
                                 mass_def=mdef)

    def sel(m, a):
        m = np.atleast_1d(m)
        a = np.atleast_1d(a)
        val = np.zeros_like(m.reshape(-1, 1) * a.reshape(1, -1))
        msk_a = (a > 0.5) & (a < 1.0)
        msk_m = (m > 1e14) & (m < 1e16)
        val[msk_m, :] += 1
        val[:, msk_a] += 1
        msk = val == 2
        val[~msk] = 0
        return val

    nc = hmc.number_counts(cosmo, selection=sel) * 4.0 * np.pi
    assert np.isfinite(nc)
    assert not np.allclose(nc, 0)


def test_hmcalculator_number_counts_zero():
    cosmo = ccl.Cosmology(
        Omega_c=0.27, Omega_b=0.045, h=0.67, sigma8=0.8, n_s=0.96,
        transfer_function='bbks', matter_power_spectrum='linear')
    mdef = ccl.halos.MassDef(200, 'matter')
    hmf = ccl.halos.MassFuncTinker10(mass_def=mdef, mass_def_strict=False)
    hbf = ccl.halos.HaloBiasTinker10(mass_def=mdef, mass_def_strict=False)

    hmc = ccl.halos.HMCalculator(mass_function=hmf, halo_bias=hbf,
                                 mass_def=mdef)

    def sel(m, a):
        m = np.atleast_1d(m)
        a = np.atleast_1d(a)
        val = np.zeros_like(m.reshape(-1, 1) * a.reshape(1, -1))
        return val

    nc = hmc.number_counts(cosmo, selection=sel) * 4.0 * np.pi
    assert np.isfinite(nc)
    assert np.allclose(nc, 0)


def test_hmcalculator_number_counts_norm():
    cosmo = ccl.Cosmology(
        Omega_c=0.27, Omega_b=0.045, h=0.67, sigma8=0.8, n_s=0.96,
        transfer_function='bbks', matter_power_spectrum='linear')
    mdef = ccl.halos.MassDef(200, 'matter')
    hmf = ccl.halos.MassFuncTinker10(mass_def=mdef, mass_def_strict=False)
    hbf = ccl.halos.HaloBiasTinker10(mass_def=mdef, mass_def_strict=False)

    hmc = ccl.halos.HMCalculator(mass_function=hmf, halo_bias=hbf,
                                 mass_def=mdef)

    def sel2(m, a):
        m = np.atleast_1d(m)
        a = np.atleast_1d(a)
        val = np.zeros_like(m.reshape(-1, 1) * a.reshape(1, -1))
        msk_a = (a > 0.5) & (a < 1.0)
        msk_m = (m > 1e14) & (m < 1e16)
        val[msk_m, :] += 1
        val[:, msk_a] += 1
        msk = val == 2
        val[~msk] = 0
        return val

    def sel4(m, a):
        m = np.atleast_1d(m)
        a = np.atleast_1d(a)
        val = np.zeros_like(m.reshape(-1, 1) * a.reshape(1, -1))
        msk_a = (a > 0.5) & (a < 1.0)
        msk_m = (m > 1e14) & (m < 1e16)
        val[msk_m, :] += 2
        val[:, msk_a] += 2
        msk = val == 4
        val[~msk] = 0
        return val

    nc2 = hmc.number_counts(cosmo, selection=sel2) * 4.0 * np.pi
    assert np.isfinite(nc2)
    assert not np.allclose(nc2, 0)
    nc4 = hmc.number_counts(cosmo, selection=sel4) * 4.0 * np.pi
    assert np.isfinite(nc4)
    assert not np.allclose(nc4, 0)

    assert np.allclose(nc2 * 2, nc4)


def test_hmcalculator_number_counts_scipy_dblquad():
    cosmo = ccl.Cosmology(
        Omega_c=0.25,
        Omega_b=0.05,
        h=0.7,
        n_s=0.96,
        sigma8=0.8,
        Omega_k=0.0,
        Omega_g=0,
        Neff=0.0,
        w0=-1.0,
        wa=0.0,
        T_CMB=2.7245,
        transfer_function='eisenstein_hu',
        matter_power_spectrum='linear'
    )
    mdef = ccl.halos.MassDef(200, 'matter')
    hmf = ccl.halos.MassFuncTinker08(mass_def=mdef, mass_def_strict=False)
    hbf = ccl.halos.HaloBiasTinker10(mass_def=mdef, mass_def_strict=False)

    amin = 0.75
    amax = 1.0
    mmin = 1e14
    mmax = 1e15

    hmc = ccl.halos.HMCalculator(
        mass_function=hmf, halo_bias=hbf, mass_def=mdef,
        log10M_min=np.log10(mmin),
        log10M_max=np.log10(mmax),
        integration_method_M='spline')

    def sel(m, a):
        m = np.atleast_1d(m)
        a = np.atleast_1d(a)
        val = np.zeros_like(m.reshape(-1, 1) * a.reshape(1, -1))
        msk_a = (a > amin) & (a < amax)
        msk_m = (m > mmin) & (m < mmax)
        val[msk_m, :] += 2
        val[:, msk_a] += 2
        msk = val == 4
        val[~msk] = 0
        val[msk] = 1.0
        return val

    def _func(m, a):
        abs_dzda = 1 / a / a
        dc = ccl.comoving_angular_distance(cosmo, a)
        ez = ccl.h_over_h0(cosmo, a)
        dh = ccl.physical_constants.CLIGHT_HMPC / cosmo['h']
        dvdz = dh * dc**2 / ez
        dvda = dvdz * abs_dzda

        val = hmf(cosmo, 10**m, a) * sel(10**m, a)
        return val[0, 0] * dvda

    mtot, _ = scipy.integrate.dblquad(
        _func,
        amin,
        amax,
        lambda x: hmc.precision['log10M_min'],
        lambda x: hmc.precision['log10M_max'],
    )

    mtot_hmc = hmc.number_counts(cosmo, selection=sel, a_min=amin, a_max=amax)
    assert np.allclose(mtot_hmc, mtot, atol=0, rtol=0.02)



import numpy as np
from colossus.cosmology import cosmology
from colossus.lss import mass_function, bias

h = 0.7
my_cosmo = {'flat': True, 'H0': 100 * h, 'Om0': 0.30, 'Ob0': 0.05, 'sigma8': 0.8, 'ns': 0.96}
cosmo_C = cosmology.setCosmology('my_cosmo', my_cosmo)

def get_mfs(model, mdef):
    m_arr=np.geomspace(1E11,1E15,9) 
    z_arr=np.array([0., 0.5, 1.])
    d_out = []
    d_out.append(m_arr)
    normfac = h**3 * np.log(10)
    for z in z_arr:
        d_out.append(mass_function.massFunction(m_arr * h, z,
                                                mdef = mdef,
                                                model = model,
                                                q_out = 'dndlnM') * normfac)
    np.savetxt("../hmf_" + model + ".txt",
               np.transpose(d_out),
               header = 'M nM(z=0) nM(z=0.5) nM(z=1)')

get_mfs('tinker08', '200m')
get_mfs('press74', 'fof')
get_mfs('sheth99', 'fof')
get_mfs('jenkins01', 'fof')
get_mfs('despali16', 'vir')
get_mfs('bocquet16', '200c')
get_mfs('watson13', '200m')
get_mfs('angulo12', 'fof')

def get_hbs(model, mdef):
    m_arr=np.geomspace(1E11,1E15,9) 
    z_arr=np.array([0., 0.5, 1.])
    d_out = []
    d_out.append(m_arr)
    normfac = h**3 * np.log(10)
    for z in z_arr:
        d_out.append(bias.haloBias(m_arr * h, z = z,
                                   mdef = mdef,
                                   model = model))
    np.savetxt("../hbf_" + model + ".txt",
               np.transpose(d_out),
               header = 'M bh(z=0) bh(z=0.5) bh(z=1)')

get_hbs('tinker10', '200m')
get_hbs('sheth01', 'fof')
get_hbs('bhattacharya11', 'fof')



__all__ = ("BaccoemuBaryons",)

import numpy as np
from copy import deepcopy

from .. import Pk2D
from . import Baryons


class BaccoemuBaryons(Baryons):
    """ The baryonic boost factor computed with the baccoemu baryons emulators.

    See `Arico et al. 2021 <https://arxiv.org/abs/2011.15018>`_ and
    https://bacco.dipc.org/emulator.html.

    .. note:: Note that masses are in units of :math:`M_\\odot`, differently
              from the original paper and baccoemu public code (where they are
              in :math:`M_\\odot/h`)

    Args:
        log10_M_c (:obj:`float`): characteristic halo mass to model baryon
                                   mass fraction (in :math:`M_\\odot`)
        log10_eta (:obj:`float`): extent of ejected gas
        log10_beta (:obj:`float`): slope of power law describing baryon mass
                                    fraction
        log10_M1_z0_cen (:obj:`float`): characteristic halo mass scale for
                                         central galaxies (in :math:`M_\\odot`)
        log10_theta_out (:obj:`float`):  outer slope of density profiles of
                                          hot gas in haloes
        log10_theta_inn (:obj:`float`): inner slope of density profiles of
                                         hot gas in haloes
        log10_M_inn (:obj:`float`): transition mass of density profiles of
                                     hot gas in haloes (in :math:`M_\\odot`)
        verbose (:obj:`bool`): Verbose output from baccoemu.
                                (default: :obj:`False`)
    """
    name = 'BaccoemuBaryons'
    __repr_attrs__ = __eq_attrs__ = ("bcm_params",)

    def __init__(self, log10_M_c=14.174, log10_eta=-0.3, log10_beta=-0.22,
                 log10_M1_z0_cen=10.674, log10_theta_out=0.25,
                 log10_theta_inn=-0.86, log10_M_inn=13.574,
                 verbose=False):
        # avoid tensorflow warnings
        import warnings
        with warnings.catch_warnings():
            warnings.filterwarnings('ignore', category=UserWarning)
            import baccoemu
            self.mpk = baccoemu.Matter_powerspectrum(verbose=verbose)
        self.a_min = self.mpk.emulator['baryon']['bounds'][-1][0]
        self.a_max = self.mpk.emulator['baryon']['bounds'][-1][1]
        self.k_min = self.mpk.emulator['baryon']['k'][0]
        self.k_max = self.mpk.emulator['baryon']['k'][-1]

        self.bcm_params = {
            'M_c': log10_M_c,
            'eta': log10_eta,
            'beta': log10_beta,
            'M1_z0_cen': log10_M1_z0_cen,
            'theta_out': log10_theta_out,
            'theta_inn': log10_theta_inn,
            'M_inn': log10_M_inn
        }

    def _sigma8tot_2_sigma8cold(self, emupars, sigma8tot):
        """Use baccoemu to convert sigma8 total matter to sigma8 cdm+baryons
        """
        if np.ndim(emupars['omega_cold']) == 1:
            _emupars = {}
            for pname in emupars:
                _emupars[pname] = emupars[pname][0]
        else:
            _emupars = emupars
        A_s_fid = 2.1e-9
        sigma8tot_fid = self.mpk.get_sigma8(cold=False,
                                            A_s=A_s_fid, **_emupars)
        A_s = (sigma8tot / sigma8tot_fid)**2 * A_s_fid
        return self.mpk.get_sigma8(cold=True, A_s=A_s, **_emupars)

    def boost_factor(self, cosmo, k, a):
        """The baccoemu BCM model boost factor for baryons.

        Args:
            cosmo (:class:`~pyccl.cosmology.Cosmology`): Cosmological parameters.
            k (:obj:`float` or `array`): Wavenumber (in :math:`{\\rm Mpc}^{-1}`).
            a (:obj:`float` or `array`): Scale factor.

        Returns:
            :obj:`float` or `array`: Correction factor to apply to \
            the power spectrum.
        """ # noqa

        # Check a ranges
        self._check_a_range(a)

        # First create the dictionary passed to baccoemu
        # if a is an array, make sure all the other parameters passed to the
        # emulator have the same len
        if np.ndim(a) == 1:
            emupars = dict(
                omega_cold=np.full((len(a)),
                                   cosmo['Omega_c'] + cosmo['Omega_b']),
                omega_baryon=np.full((len(a)), cosmo['Omega_b']),
                ns=np.full((len(a)), cosmo['n_s']),
                hubble=np.full((len(a)), cosmo['h']),
                neutrino_mass=np.full((len(a)), np.sum(cosmo['m_nu'])),
                w0=np.full((len(a)), cosmo['w0']),
                wa=np.full((len(a)), cosmo['wa']),
                expfactor=a
            )
        else:
            emupars = dict(
                omega_cold=cosmo['Omega_c'] + cosmo['Omega_b'],
                omega_baryon=cosmo['Omega_b'],
                ns=cosmo['n_s'],
                hubble=cosmo['h'],
                neutrino_mass=np.sum(cosmo['m_nu']),
                w0=cosmo['w0'],
                wa=cosmo['wa'],
                expfactor=a
            )

        # if cosmo contains sigma8, we use it for baccoemu, otherwise we pass
        # A_s to the emulator
        if np.isnan(cosmo['A_s']):
            # note that ccl parametrises sigma8 of the total matter power
            # spectrum while baccoemu defines it in terms of the cdm+baryons
            # power spectrum; so we have to convert from total to cold sigma8
            sigma8tot = cosmo['sigma8']
            sigma8cold = self._sigma8tot_2_sigma8cold(emupars, sigma8tot)
            if np.ndim(a) == 1:
                emupars['sigma8_cold'] = np.full((len(a)), sigma8cold)
            else:
                emupars['sigma8_cold'] = sigma8cold
        else:
            if np.ndim(a) == 1:
                emupars['A_s'] = np.full((len(a)), cosmo['A_s'])
            else:
                emupars['A_s'] = cosmo['A_s']

        # change masses from Msun to Msun/h
        _bcm_params = deepcopy(self.bcm_params)
        l10h = np.log10(emupars['hubble'])
        for key in ['M_c', 'M1_z0_cen', 'M_inn']:
            _bcm_params[key] += l10h

        # baccoemu internally interpolates k with a cubic spline
        # it returns k, boost, so, since we are already requesting a specific
        # k-vector we can ignore the first returned object
        _, fka = self.mpk.get_baryonic_boost(k=k / cosmo['h'],
                                             **{**emupars, **_bcm_params})

        return fka

    def update_parameters(self, log10_M_c=None, log10_eta=None,
                          log10_beta=None, log10_M1_z0_cen=None,
                          log10_theta_out=None, log10_theta_inn=None,
                          log10_M_inn=None):
        """Update parameters. All parameters set to ``None`` will
        be left untouched.

        Args:
            log10_M_c (:obj:`float`): characteristic halo mass to model baryon
                mass fraction (in :math:`M_\\odot`)
            log10_eta (:obj:`float`): extent of ejected gas
            log10_beta (:obj:`float`): slope of power law describing baryon
                mass fraction
            log10_M1_z0_cen (:obj:`float`): characteristic halo mass scale for
                central galaxies (in :math:`M_\\odot`)
            log10_theta_out (:obj:`float`):  outer slope of density profiles of
                hot gas in haloes
            log10_theta_inn (:obj:`float`): inner slope of density profiles of
                hot gas in haloes
            log10_M_inn (:obj:`float`): transition mass of density profiles of
                hot gas in haloes (in :math:`M_\\odot`)
        """
        _kwargs = locals()
        _new_bcm_params = {key: _kwargs[key] for key in
                           set(list(_kwargs.keys())) - set(['self'])}
        new_bcm_params = {}
        for key in _new_bcm_params:
            if _new_bcm_params[key] is not None:
                new_bcm_params[key[6:]] = _new_bcm_params[key]
        self.bcm_params.update(new_bcm_params)

    def _include_baryonic_effects(self, cosmo, pk):
        # Applies boost factor
        a_arr, lk_arr, pk_arr = pk.get_spline_arrays()
        k_arr = np.exp(lk_arr)
        fka = self.boost_factor(cosmo, k_arr, a_arr)
        pk_arr *= fka

        if pk.psp.is_log:
            np.log(pk_arr, out=pk_arr)  # in-place log

        return Pk2D(a_arr=a_arr, lk_arr=lk_arr, pk_arr=pk_arr,
                    is_logp=pk.psp.is_log,
                    extrap_order_lok=pk.extrap_order_lok,
                    extrap_order_hik=pk.extrap_order_hik)

    def _check_a_range(self, a):
        if np.ndim(a) == 0:
            a_min, a_max = a, a
        else:
            a_min = min(a)
            a_max = max(a)
        if a_min < self.a_min or a_max > self.a_max:
            raise ValueError(f"Requested scale factor outside the bounds of "
                             f"the emulator: {(a_min, a_max)} outside of "
                             f"{((self.a_min, self.a_max))}")



__all__ = ("ConcentrationDiemer15",)

import numpy as np

from . import Concentration, get_delta_c


class ConcentrationDiemer15(Concentration):
    """Concentration-mass relation by `Diemer & Kravtsov 2015
    <https://arxiv.org/abs/1407.4730>`_. This parametrization
    is only valid for S.O. masses with :math:`\\Delta = 200`
    times the critical density.

    Args:
        mass_def (:class:`~pyccl.halos.massdef.MassDef` or :obj:`str`):
            a mass definition object, or a name string.
    """
    name = 'Diemer15'

    def __init__(self, *, mass_def="200c"):
        super().__init__(mass_def=mass_def)

    def _setup(self):
        self.kappa = 1.0
        self.phi_0 = 6.58
        self.phi_1 = 1.27
        self.eta_0 = 7.28
        self.eta_1 = 1.56
        self.alpha = 1.08
        self.beta = 1.77

    def _check_mass_def_strict(self, mass_def):
        return mass_def.name != "200c"

    def _concentration(self, cosmo, M, a):
        # Compute power spectrum slope
        R = cosmo.mass2radius_lagrangian(M)
        k_R = 2.0 * np.pi / R * self.kappa

        cosmo.compute_linear_power()
        pk = cosmo.get_linear_power()
        n = pk(k_R, a, derivative=True)

        sig = cosmo.sigmaM(M, a)
        delta_c = get_delta_c(cosmo, a, kind='EdS')
        nu = delta_c / sig

        floor = self.phi_0 + n * self.phi_1
        nu0 = self.eta_0 + n * self.eta_1
        c = 0.5 * floor * ((nu0 / nu)**self.alpha +
                           (nu / nu0)**self.beta)
        return c



import numpy as np
import pyccl as ccl
import pytest
from timeit import default_timer

ccl.gsl_params.LENSING_KERNEL_SPLINE_INTEGRATION = False
COSMO = ccl.Cosmology(
    Omega_c=0.27, Omega_b=0.045, h=0.67, sigma8=0.8, n_s=0.96,
    transfer_function='bbks', matter_power_spectrum='linear')


@pytest.mark.parametrize('method', ['bessel', 'legendre', 'fftlog'])
def test_correlation_smoke(method):
    z = np.linspace(0., 1., 200)
    n = np.ones(z.shape)
    lens = ccl.WeakLensingTracer(COSMO, dndz=(z, n))

    ell = np.logspace(1, 3, 5)
    cl = ccl.angular_cl(COSMO, lens, lens, ell)

    t_arr = np.logspace(-2., np.log10(5.), 5)
    t_lst = [t for t in t_arr]
    t_scl = 2.
    t_int = 2

    for tval in [t_arr, t_lst, t_scl, t_int]:
        corr = ccl.correlation(
            COSMO, ell=ell, C_ell=cl, theta=tval, type='NN', method=method)
        assert np.all(np.isfinite(corr))
        assert np.shape(corr) == np.shape(tval)


@pytest.mark.parametrize(
    'rval',
    [50,
     50.0,
     np.logspace(1, 2, 5),
     [r for r in np.logspace(1, 2, 5)]])
def test_correlation_3d_smoke(rval):
    a = 0.8
    corr = ccl.correlation_3d(COSMO, a=a, r=rval)
    assert np.all(np.isfinite(corr))
    assert np.shape(corr) == np.shape(rval)


@pytest.mark.parametrize(
    'sval',
    [50,
     50.0,
     np.logspace(1, 2, 5),
     [r for r in np.logspace(1, 2, 5)]])
def test_correlation_3dRSD_smoke(sval):
    a = 0.8
    mu = 0.7
    beta = 0.5
    corr = ccl.correlation_3dRsd(COSMO, a=a, r=sval, mu=mu, beta=beta)
    assert np.all(np.isfinite(corr))
    assert np.shape(corr) == np.shape(sval)


@pytest.mark.parametrize(
    'sval',
    [50,
     50.0,
     np.logspace(1, 2, 5),
     [r for r in np.logspace(1, 2, 5)]])
def test_correlation_3dRSD_avgmu_smoke(sval):
    a = 0.8
    beta = 0.5
    corr = ccl.correlation_3dRsd_avgmu(COSMO, a=a, r=sval, beta=beta)
    assert np.all(np.isfinite(corr))
    assert np.shape(corr) == np.shape(sval)


@pytest.mark.parametrize('l', [0, 2, 4])
@pytest.mark.parametrize(
    'sval',
    [50,
     50.0,
     np.logspace(1, 2, 5),
     [r for r in np.logspace(1, 2, 5)]])
def test_correlation_3dRSD_multipole_smoke(sval, l):
    a = 0.8
    beta = 0.5
    corr = ccl.correlation_multipole(COSMO, a=a, beta=beta,
                                     ell=l, r=sval)
    assert np.all(np.isfinite(corr))
    assert np.shape(corr) == np.shape(sval)


@pytest.mark.parametrize(
    'sval',
    [50,
     50.0,
     np.logspace(1, 2, 5),
     [r for r in np.logspace(1, 2, 5)]])
def test_correlation_pi_sigma_smoke(sval):
    a = 0.8
    beta = 0.5
    pie = 1
    corr = ccl.correlation_pi_sigma(COSMO, a=a, beta=beta, pi=pie, sigma=sval)
    assert np.all(np.isfinite(corr))
    assert np.shape(corr) == np.shape(sval)


def test_correlation_raises():
    with pytest.raises(ValueError):
        ccl.correlation(COSMO, ell=[1], C_ell=[1e-3], theta=[1], method='blah')
    with pytest.raises(ValueError):
        ccl.correlation(COSMO, ell=[1], C_ell=[1e-3], theta=[1], type='blah')


def test_correlation_zero():
    ell = np.arange(2, 100000)
    C_ell = np.zeros(ell.size)
    theta = np.logspace(0, 2, 10000)
    t0 = default_timer()
    corr = ccl.correlation(COSMO, ell=ell, C_ell=C_ell, theta=theta)
    t1 = default_timer()
    # if the short-cut has worked this should take
    # less than 1 second at the absolute outside
    assert t1 - t0 < 1.0
    assert (corr == np.zeros(theta.size)).all()


def test_correlation_zero_ends():
    # This should give an error instead of crashing
    ell = np.arange(2, 1001)
    C_ell = np.zeros(ell.size)
    C_ell[500] = 1.0
    theta = np.logspace(0, 2, 20)
    with pytest.raises(ccl.CCLError):
        ccl.correlation(COSMO, ell=ell, C_ell=C_ell, theta=theta)


ccl.gsl_params.reload()  # reset to the default parameters



import numpy as np
import pyccl as ccl
import pytest

cemu_iv = ccl.CosmicemuMTIVPk('tot')
cemu_ii = ccl.CosmicemuMTIIPk('tot')


@pytest.mark.parametrize('cemu', [cemu_ii, cemu_iv])
def test_cosmicemu_smoke(cemu):
    cosmo = ccl.CosmologyVanillaLCDM(matter_power_spectrum=cemu)
    k, pk = cemu.get_pk_at_a(cosmo, 1.0)
    pk2 = cosmo.nonlin_matter_power(k, 1.0)
    assert np.allclose(pk, pk2, atol=0, rtol=1E-6)


@pytest.mark.parametrize('cemu', [cemu_ii, cemu_iv])
def test_cosmicemu_As_raises(cemu):
    cosmo = ccl.Cosmology(Omega_c=0.25,
                          Omega_b=0.05,
                          h=0.67, n_s=0.96,
                          A_s=2E-9)
    # CosmicEmu needs sigma8
    with pytest.raises(ValueError):
        cemu.get_pk_at_a(cosmo, 1.0)


@pytest.mark.parametrize('cemu', [cemu_ii, cemu_iv])
def test_cosmicemu_outbound(cemu):
    cosmo = ccl.Cosmology(Omega_c=0.25,
                          Omega_b=0.05,
                          h=0.67, n_s=0.96,
                          sigma8=1.5)
    # sigma8 out of bounds
    with pytest.raises(ValueError):
        cemu.get_pk_at_a(cosmo, 1.0)



import pyccl as ccl
from pyccl.modified_gravity import MuSigmaMG, ModifiedGravity

import pytest


@pytest.mark.parametrize('tf', ['eisenstein_hu', 'bbks'])
def test_mu_sigma_transfer_err(tf):
    with pytest.raises(ccl.CCLError):
        cosmo = ccl.Cosmology(
            Omega_c=0.25,
            Omega_b=0.05,
            h=0.7,
            A_s=2.1e-9,
            n_s=0.96,
            mg_parametrization=MuSigmaMG(mu_0=0.1, sigma_0=0.2),
            transfer_function=tf,
            matter_power_spectrum='linear'
        )
        ccl.linear_matter_power(cosmo, 1, 1)


def test_mg_error():
    class NotMG:
        pass

    class NotMuSigma(ModifiedGravity):
        pass

    with pytest.raises(ValueError):
        ccl.Cosmology(
            Omega_c=0.25,
            Omega_b=0.05,
            h=0.7,
            sigma8=0.8,
            n_s=0.96,
            mg_parametrization=NotMG(),
            transfer_function="bbks",
        )

    with pytest.raises(NotImplementedError):
        ccl.Cosmology(
            Omega_c=0.25,
            Omega_b=0.05,
            h=0.7,
            sigma8=0.8,
            n_s=0.96,
            mg_parametrization=NotMuSigma(),
            transfer_function="bbks",
        )



import numpy as np
import pyccl as ccl
import pytest

# Set tolerances
BOOST_TOLERANCE = 1e-5

# Set up the cosmological parameters to be used
cosmo = ccl.Cosmology(Omega_c=0.27, Omega_b=0.045, h=0.67,
                      sigma8=0.8, n_s=0.96,
                      transfer_function='eisenstein_hu')
k = 0.5*cosmo['h']
a = 1.

# Set up power without baryons
pk2D_no_baryons = cosmo.get_nonlin_power()

fbarcvec = np.linspace(0.25, 1, 20)
mdef_vec = ['500c', '200c']


def compare_boost():
    vdboost = []
    vdboost_expect = []
    pk_nl = pk2D_no_baryons(k, a)
    for mdef in mdef_vec:
        for f in fbarcvec:
            # Takes ftilde as argument
            vD19 = ccl.BaryonsvanDaalen19(fbar=f, mass_def=mdef)
            pk2D_with_baryons = vD19.include_baryonic_effects(
                cosmo, pk2D_no_baryons)
            # Takes k in units of 1/Mpc as argument
            pk_nl_bar = pk2D_with_baryons(k, a)
            vdboost.append(pk_nl_bar/pk_nl-1)
            if mdef == '500c':
                vdboost_expect.append(-np.exp(-5.99*f-0.5107))
            else:
                vdboost_expect.append(-np.exp(-5.816*f-0.4005))
    assert np.allclose(
        vdboost, vdboost_expect, atol=1e-5, rtol=BOOST_TOLERANCE)


def test_boost_model():
    compare_boost()


def test_baryons_from_name():
    baryons = ccl.BaryonsvanDaalen19(fbar=0.7, mass_def='500c')
    bar2 = ccl.Baryons.from_name('vanDaalen19')
    assert baryons.name == bar2.name
    assert baryons.name == 'vanDaalen19'


def test_baryons_vd19_raises():
    with pytest.raises(ValueError):
        ccl.BaryonsvanDaalen19(fbar=0.7, mass_def='blah')
    b = ccl.BaryonsvanDaalen19(fbar=0.7, mass_def='500c')
    with pytest.raises(ValueError):
        b.update_parameters(mass_def='blah')


def test_update_params():
    b = ccl.BaryonsvanDaalen19(fbar=0.7, mass_def='500c')
    b.update_parameters(fbar=0.6, mass_def='200c')
    assert b.mass_def == '200c'
    assert b.fbar == 0.6



import numpy as np
import pyccl as ccl
import pytest

CEMU_TOL = 1E-3
# Old emulator tolerance
EMU_TOLERANCE = 3.0E-2
cemu_old = ccl.CosmicemuMTIIPk('tot')


@pytest.mark.parametrize("kind", ["tot", "cb"])
def test_cemu_mtiv(kind):
    cemu = ccl.CosmicemuMTIVPk(kind)
    for i, omega_nu in enumerate([0.0, 1E-3]):
        m_nu = ccl.nu_masses(Omega_nu_h2=omega_nu, mass_split='equal')
        h = 0.67
        cosmo = ccl.Cosmology(
            Omega_c=0.3-0.05-omega_nu/h**2,
            Omega_b=0.05,
            h=h,
            sigma8=0.8,
            n_s=0.96,
            m_nu=m_nu,
            mass_split='equal',
            w0=-1.0,
            wa=0.0,
            matter_power_spectrum=cemu)

        zs = np.array([0.0, 1.0])
        for iz, a in enumerate(1/(1+zs)):
            k, pk = np.loadtxt(
                f"./benchmarks/data/cosmo{i+1}_{kind}_{iz}.txt",
                unpack=True)
            pkh = cosmo.nonlin_matter_power(k, a)
            assert np.allclose(pk, pkh, atol=0, rtol=CEMU_TOL)

            ktest, pktest = cemu.get_pk_at_a(cosmo, a)
            pkh = np.exp(np.interp(np.log(k), np.log(ktest), np.log(pktest)))
            assert np.allclose(pk, pkh, atol=0, rtol=CEMU_TOL)


@pytest.mark.parametrize("kind", ["tot", "cb"])
def test_cemu_mtii(kind):
    cemu = ccl.CosmicemuMTIIPk(kind)
    for i, omega_nu in enumerate([0.0, 1E-3]):
        m_nu = ccl.nu_masses(Omega_nu_h2=omega_nu, mass_split='equal')
        h = 0.67
        cosmo = ccl.Cosmology(
            Omega_c=0.3-0.05-omega_nu/h**2,
            Omega_b=0.05,
            h=h,
            sigma8=0.8,
            n_s=0.96,
            m_nu=m_nu,
            mass_split='equal',
            w0=-1.0,
            wa=0.0,
            matter_power_spectrum=cemu)

        zs = np.array([0.0, 1.0])
        for iz, a in enumerate(1/(1+zs)):
            k, pk = np.loadtxt(
                f"./benchmarks/data/cosmo{i+1}_MTII_{kind}_{iz}.txt",
                unpack=True)
            pkh = cosmo.nonlin_matter_power(k, a)
            assert np.allclose(pk, pkh, atol=0, rtol=CEMU_TOL)

            ktest, pktest = cemu.get_pk_at_a(cosmo, a)
            pkh = np.exp(np.interp(np.log(k), np.log(ktest), np.log(pktest)))
            assert np.allclose(pk, pkh, atol=0, rtol=CEMU_TOL)


@pytest.mark.parametrize('model', list(range(4)))
def test_emu_nu_old(model):
    model_numbers = [38, 39, 40, 42]
    data = np.loadtxt(
        "./benchmarks/data/emu_nu_smooth_pk_M%d.txt" % model_numbers[model])

    cosmos = np.loadtxt("./benchmarks/data/emu_nu_cosmologies.txt")

    mnu = ccl.nu_masses(Omega_nu_h2=cosmos[model, 7]*cosmos[model, 2]**2,
                        mass_split='equal')

    cosmo = ccl.Cosmology(
        Omega_c=cosmos[model, 0],
        Omega_b=cosmos[model, 1],
        h=cosmos[model, 2],
        sigma8=cosmos[model, 3],
        n_s=cosmos[model, 4],
        w0=cosmos[model, 5],
        wa=cosmos[model, 6],
        m_nu=mnu,
        mass_split='list',
        Neff=3.04,
        Omega_g=0,
        Omega_k=0,
        transfer_function='boltzmann_camb',
        matter_power_spectrum=cemu_old,
    )

    a = 1
    k = data[:, 0]

    pk = ccl.nonlin_matter_power(cosmo, k, a)

    err = np.abs(pk/data[:, 1]-1)
    assert np.allclose(err, 0, rtol=0, atol=EMU_TOLERANCE)


@pytest.mark.parametrize('model', [0, 1, 3, 5])
def test_emu_old(model):
    model_numbers = [1, 3, 5, 6, 8, 10]
    data = np.loadtxt(
        "./benchmarks/data/emu_smooth_pk_M%d.txt" % model_numbers[model])

    cosmos = np.loadtxt("./benchmarks/data/emu_cosmologies.txt")

    cosmo = ccl.Cosmology(
        Omega_c=cosmos[model, 0],
        Omega_b=cosmos[model, 1],
        h=cosmos[model, 2],
        sigma8=cosmos[model, 3],
        n_s=cosmos[model, 4],
        w0=cosmos[model, 5],
        wa=cosmos[model, 6],
        Neff=3.04,
        Omega_g=0,
        Omega_k=0,
        transfer_function='bbks',
        matter_power_spectrum=cemu_old,
    )

    a = 1
    k = data[:, 0]
    pk = ccl.nonlin_matter_power(cosmo, k, a)
    err = np.abs(pk/data[:, 1]-1)
    assert np.allclose(err, 0, rtol=0, atol=EMU_TOLERANCE)


@pytest.mark.parametrize('model', list(range(3)))
def test_emu_lin_old(model):
    cosmos = np.loadtxt("./benchmarks/data/emu_input_cosmologies.txt")

    mnu = ccl.nu_masses(Omega_nu_h2=cosmos[model, 7]*cosmos[model, 2]**2,
                        mass_split='equal')

    cosmo = ccl.Cosmology(
        Omega_c=cosmos[model, 0],
        Omega_b=cosmos[model, 1],
        h=cosmos[model, 2],
        sigma8=cosmos[model, 3],
        n_s=cosmos[model, 4],
        w0=cosmos[model, 5],
        wa=cosmos[model, 6],
        m_nu=mnu,
        mass_split='list',
        Neff=3.04,
        Omega_g=0,
        Omega_k=0,
        transfer_function='boltzmann_camb',
        matter_power_spectrum=cemu_old,
    )

    a = 1
    k = np.logspace(-3, -2, 50)

    pk = ccl.nonlin_matter_power(cosmo, k, a)
    pk_lin = ccl.linear_matter_power(cosmo, k, a)

    err = np.abs(pk/pk_lin-1)
    assert np.allclose(err, 0, rtol=0, atol=EMU_TOLERANCE)



import numpy as np
import pytest
from unittest import mock
import pyccl as ccl
import sys
from pyccl.boltzmann import get_isitgr_pk_lin
from pyccl.modified_gravity import MuSigmaMG

try:
    from importlib import reload
except ImportError:
    pass  # in 2.7, reload is a global function.


@pytest.mark.parametrize('tf', [
    'bbks', 'eisenstein_hu', 'boltzmann_class', 'boltzmann_camb',
    'boltzmann_isitgr'])
def test_power_mu_sigma_sigma8norm(tf):
    cosmo = ccl.Cosmology(
        Omega_c=0.27, Omega_b=0.045, h=0.67, sigma8=0.8, n_s=0.96,
        transfer_function=tf)

    cosmo_musig = ccl.Cosmology(
        Omega_c=0.27, Omega_b=0.045, h=0.67, sigma8=0.8, n_s=0.96,
        transfer_function=tf,
        mg_parametrization=MuSigmaMG(mu_0=0.1, sigma_0=0.2))

    # make sure sigma8 is correct
    assert np.allclose(ccl.sigma8(cosmo_musig), 0.8)

    if tf != 'boltzmann_isitgr':
        # make sure P(k) ratio is right
        a = 0.8
        gfac = (
            ccl.growth_factor(cosmo, a) / ccl.growth_factor(cosmo_musig, a))**2
        pk_rat = (
            ccl.linear_matter_power(cosmo, 1e-4, a) /
            ccl.linear_matter_power(cosmo_musig, 1e-4, a))
        assert np.allclose(pk_rat, gfac)

    with mock.patch.dict(sys.modules, {'isitgr': None}):
        with pytest.raises(ModuleNotFoundError):
            get_isitgr_pk_lin(cosmo)
    # Importing ccl without isitgr is fine.  No ModuleNotFoundError triggered.
    with mock.patch.dict(sys.modules, {'isitgr': None}):
        reload(ccl.boltzmann)


@pytest.mark.parametrize('tf', [
    'boltzmann_class', 'boltzmann_camb', 'boltzmann_isitgr'])
def test_power_mu_sigma_sigma8norm_norms_consistent(tf):
    # make a cosmo with A_s
    cosmo = ccl.Cosmology(
        Omega_c=0.27, Omega_b=0.045, h=0.67, A_s=2e-9, n_s=0.96,
        transfer_function=tf,
        mg_parametrization=MuSigmaMG(mu_0=0.1, sigma_0=0.2))
    sigma8 = ccl.sigma8(cosmo)

    # remake same but now give sigma8
    cosmo_s8 = ccl.Cosmology(
        Omega_c=0.27, Omega_b=0.045, h=0.67, sigma8=sigma8, n_s=0.96,
        transfer_function=tf,
        mg_parametrization=MuSigmaMG(mu_0=0.1, sigma_0=0.2))

    # make sure they come out the same-ish
    assert np.allclose(ccl.sigma8(cosmo), ccl.sigma8(cosmo_s8))

    if tf != 'boltzmann_isitgr':
        # and that the power spectra look right
        a = 0.8
        gfac = (
            ccl.growth_factor(cosmo, a) / ccl.growth_factor(cosmo_s8, a))**2
        pk_rat = (
            ccl.linear_matter_power(cosmo, 1e-4, a) /
            ccl.linear_matter_power(cosmo_s8, 1e-4, a))
        assert np.allclose(pk_rat, gfac)


def test_nonlin_camb_MG_error():
    Omega_c = 0.25
    Omega_b = 0.05
    n_s = 0.97
    h = 0.7

    ccl_cosmo = ccl.Cosmology(Omega_c=Omega_c, Omega_b=Omega_b, h=h, m_nu=0.0,
                              A_s=2.1e-9, n_s=n_s,
                              transfer_function="boltzmann_camb",
                              matter_power_spectrum="camb",
                              mg_parametrization=MuSigmaMG(
                                  mu_0=0.1, sigma_0=0.2))

    k = np.logspace(-3, 1, 10)

    # Check that non-linear power spectrum isn't being used with sigma8
    with pytest.raises(ValueError):
        ccl.nonlin_matter_power(ccl_cosmo, k, 1.0)



import numpy as np
import pytest
import pyccl as ccl
from .test_cclobject import check_eq_repr_hash


COSMO = ccl.Cosmology(
    Omega_c=0.27, Omega_b=0.045, h=0.67, sigma8=0.8, n_s=0.96,
    transfer_function='bbks', matter_power_spectrum='linear')
CONCS = [ccl.halos.ConcentrationDiemer15,
         ccl.halos.ConcentrationBhattacharya13,
         ccl.halos.ConcentrationPrada12,
         ccl.halos.ConcentrationKlypin11,
         ccl.halos.ConcentrationDuffy08,
         ccl.halos.ConcentrationIshiyama21,
         ccl.halos.ConcentrationConstant]
MS = [1E13, [1E12, 1E15], np.array([1E12, 1E15])]
# None of our concentration-mass relations
# are defined for FoF halos, or 400 critical.
MDEF = ccl.halos.MassDefFof
M400 = ccl.halos.MassDef(400, 'critical')


def test_Concentration_eq_repr_hash():
    # Test eq, repr, hash for Concentration.
    CM1 = ccl.halos.Concentration.from_name("Duffy08")()
    CM2 = ccl.halos.ConcentrationDuffy08()
    assert check_eq_repr_hash(CM1.mass_def, CM2.mass_def)
    assert check_eq_repr_hash(CM1, CM2)

    M200m = ccl.halos.MassDef200m
    CM3 = ccl.halos.ConcentrationDuffy08(mass_def=M200m)
    assert check_eq_repr_hash(CM1.mass_def, CM3.mass_def, equal=False)
    assert check_eq_repr_hash(CM1, CM3, equal=False)


@pytest.mark.parametrize('cM_class', CONCS)
def test_cM_subclasses_smoke(cM_class):
    cM = cM_class()
    for m in MS:
        c = cM(COSMO, m, 0.9)
        assert np.all(np.isfinite(c))
        assert np.shape(c) == np.shape(m)


def test_cM_duffy_smoke():
    md = ccl.halos.MassDef('vir', 'critical')
    cM = ccl.halos.ConcentrationDuffy08(mass_def=md)
    for m in MS:
        c = cM(COSMO, m, 0.9)
        assert np.all(np.isfinite(c))
        assert np.shape(c) == np.shape(m)


@pytest.mark.parametrize('cM_class', CONCS[:-1])
def test_cM_mdef_raises(cM_class):
    # testing strings
    with pytest.raises(ValueError):
        cM_class(mass_def=MDEF)
    # testing numbers
    with pytest.raises(ValueError):
        cM_class(mass_def=M400)

    if cM_class.name == "Ishiyama21":
        with pytest.raises(ValueError):
            cM_class(mass_def="500c", Vmax=True)

        with pytest.raises(ValueError):
            cM_class(mass_def="200m")


@pytest.mark.parametrize('name', ['Duffy08', 'Diemer15'])
def test_cM_from_string(name):
    cM_class = ccl.halos.Concentration.from_name(name)
    assert cM_class == ccl.halos.Concentration.from_name(name)
    cM = cM_class()
    for m in MS:
        c = cM(COSMO, m, 0.9)
        assert np.all(np.isfinite(c))
        assert np.shape(c) == np.shape(m)


def test_cM_from_string_raises():
    with pytest.raises(KeyError):
        ccl.halos.Concentration.from_name('Duffy09')



import numpy as np
import pyccl as ccl


def test_hmcalculator_number_counts_numcosmo():
    cosmo = ccl.Cosmology(
        Omega_c=0.25,
        Omega_b=0.05,
        h=0.7,
        n_s=0.96,
        sigma8=0.8,
        Omega_k=0.0,
        Omega_g=0,
        Neff=0.0,
        w0=-1.0,
        wa=0.0,
        T_CMB=2.7245,
        transfer_function='eisenstein_hu',
        matter_power_spectrum='linear'
    )
    mdef = ccl.halos.MassDef(200, 'matter')
    hmf = ccl.halos.MassFuncTinker08(mass_def=mdef, mass_def_strict=False)
    hbf = ccl.halos.HaloBiasTinker10(mass_def=mdef, mass_def_strict=False)

    benches = np.loadtxt("./benchmarks/data/numcosmo_cluster_counts.txt")

    for i in range(benches.shape[0]):
        bench = benches[i, :]
        hmc = ccl.halos.HMCalculator(
            mass_function=hmf, halo_bias=hbf, mass_def=mdef,
            log10M_min=np.log10(bench[1]),
            log10M_max=np.log10(bench[2]),
            integration_method_M='spline')

        a_2 = 1.0 / (1.0 + bench[4])
        a_1 = 1.0 / (1.0 + bench[3])

        def sel(m, a):
            m = np.atleast_1d(m)
            a = np.atleast_1d(a)
            val = np.zeros_like(m.reshape(-1, 1) * a.reshape(1, -1))
            msk_a = (a > a_2) & (a < a_1)
            msk_m = (m > bench[1]) & (m < bench[2])
            val[msk_m, :] += 1
            val[:, msk_a] += 1
            msk = val == 2
            val[~msk] = 0
            val[msk] = 1.0
            return val

        area = 200 * (np.pi / 180)**2

        nc = hmc.number_counts(cosmo, selection=sel,
                               a_min=a_2, a_max=a_1) * area
        assert np.isfinite(nc)
        assert not np.allclose(nc, 0)

        tol = max(0.013, np.sqrt(bench[0]) / bench[0] / 10)
        print(nc, bench[0], nc/bench[0]-1, tol)
        assert np.allclose(nc, bench[0], atol=0, rtol=tol)


def test_hmcalculator_number_counts_numcosmo_highacc():
    cosmo = ccl.Cosmology(
        Omega_c=0.25,
        Omega_b=0.05,
        h=0.7,
        n_s=0.96,
        sigma8=0.8,
        Omega_k=0.0,
        Omega_g=0,
        Neff=0.0,
        w0=-1.0,
        wa=0.0,
        T_CMB=2.7245,
        transfer_function='eisenstein_hu',
        matter_power_spectrum='linear'
    )
    mdef = ccl.halos.MassDef(200, 'matter')
    hmf = ccl.halos.MassFuncTinker08(mass_def=mdef, mass_def_strict=False)
    hbf = ccl.halos.HaloBiasTinker10(mass_def=mdef, mass_def_strict=False)

    benches = np.loadtxt("./benchmarks/data/numcosmo_cluster_counts.txt")

    for i in range(benches.shape[0]):
        bench = benches[i, :]
        hmc = ccl.halos.HMCalculator(
            mass_function=hmf, halo_bias=hbf, mass_def=mdef,
            log10M_min=np.log10(bench[1]),
            log10M_max=np.log10(bench[2]),
            integration_method_M='spline',
            nM=4096,
        )

        a_2 = 1.0 / (1.0 + bench[4])
        a_1 = 1.0 / (1.0 + bench[3])

        def sel(m, a):
            m = np.atleast_1d(m)
            a = np.atleast_1d(a)
            val = np.zeros_like(m.reshape(-1, 1) * a.reshape(1, -1))
            msk_a = (a > a_2) & (a < a_1)
            msk_m = (m > bench[1]) & (m < bench[2])
            val[msk_m, :] += 1
            val[:, msk_a] += 1
            msk = val == 2
            val[~msk] = 0
            val[msk] = 1.0
            return val

        area = 200 * (np.pi / 180)**2

        nc = hmc.number_counts(
            cosmo, selection=sel,
            a_min=a_2,
            a_max=a_1,
            na=4096,
        ) * area
        assert np.isfinite(nc)
        assert not np.allclose(nc, 0)

        tol = 1e-3
        print(nc, bench[0], nc/bench[0]-1, tol)
        assert np.allclose(nc, bench[0], atol=0, rtol=tol)



import numpy as np
import pyccl as ccl

import pytest


# FIXME: these are not real standards
# tolerence on abs difference in r^2 xi(r) for the range
# r = 0.1 - 100 Mpc (40 points in r) for z=0,1,2,3,4,5
CORR_TOLERANCE1 = 0.10
# tolerence on abs difference in r^2 xi(r) for the range
# r = 50 - 250 Mpc (100 points in r) for z=0,1,2,3,4,5
CORR_TOLERANCE2 = 0.10


@pytest.mark.parametrize('model', list(range(3)))
def test_correlation_3drsd(model):
    Omega_v = [0.7, 0.7, 0.7, 0.65, 0.75]
    w_0 = [-1.0, -0.9, -0.9, -0.9, -0.9]
    w_a = [0.0, 0.0, 0.1, 0.1, 0.1]
    beta = [0.512796247727676, 0.5107254761543404, 0.5102500557881973]

    cosmo = ccl.Cosmology(
        Omega_c=0.25,
        Omega_b=0.05,
        h=0.7,
        sigma8=0.8,
        n_s=0.96,
        Neff=3.046,
        mass_split='normal',
        Omega_g=0,
        Omega_k=1.0 - 0.25 - 0.05 - Omega_v[model],
        w0=w_0[model],
        wa=w_a[model],
        transfer_function='boltzmann_class',
        matter_power_spectrum='halofit')

    data = np.loadtxt("./benchmarks/data/model%d_xiRSD.txt" % (model+1))
    N1 = 40
    data1 = data[:N1, :]
    r1 = data1[:, 0]
    data2 = data[N1:, :]
    r2 = data2[:, 0]
    a = 1.0

    xi1 = ccl.correlation_3dRsd_avgmu(cosmo, a=a, r=r1, beta=beta[model])
    err = np.abs(r1*r1*(xi1-data1[:, 1]))
    assert np.allclose(err, 0, rtol=0, atol=CORR_TOLERANCE1)

    xi2 = ccl.correlation_3dRsd_avgmu(cosmo, a=a, r=r2, beta=beta[model])
    err = np.abs(r2*r2*(xi2-data2[:, 1]))
    assert np.allclose(err, 0, rtol=0, atol=CORR_TOLERANCE2)



import numpy as np
import pyccl as ccl

import pytest

HALOMOD_TOLERANCE = 1E-3


@pytest.mark.parametrize('model', list(range(3)))
def test_halomod(model):
    Omega_c = [0.2500, 0.2265, 0.2685]
    Omega_b = [0.0500, 0.0455, 0.0490]
    h = [0.7000, 0.7040, 0.6711]
    sigma8 = [0.8000, 0.8100, 0.8340]
    n_s = [0.9600, 0.9670, 0.9624]

    cosmo = ccl.Cosmology(
        Omega_c=Omega_c[model],
        Omega_b=Omega_b[model],
        h=h[model],
        sigma8=sigma8[model],
        n_s=n_s[model],
        Neff=0,
        m_nu=0,
        Omega_k=0,
        transfer_function='eisenstein_hu',
        matter_power_spectrum='linear')

    data_z0 = np.loadtxt("./benchmarks/data/pk_hm_c%d_z0.txt" % (model+1))
    data_z1 = np.loadtxt("./benchmarks/data/pk_hm_c%d_z1.txt" % (model+1))

    mass_def = ccl.halos.MassDef('vir', 'matter')
    hmf = ccl.halos.MassFuncSheth99(mass_def=mass_def,
                                    mass_def_strict=False,
                                    use_delta_c_fit=True)
    hbf = ccl.halos.HaloBiasSheth99(mass_def=mass_def,
                                    mass_def_strict=False)
    cM = ccl.halos.ConcentrationDuffy08(mass_def=mass_def)
    prf = ccl.halos.HaloProfileNFW(mass_def=mass_def, concentration=cM)
    hmc = ccl.halos.HMCalculator(mass_function=hmf, halo_bias=hbf,
                                 mass_def=mass_def)

    z = 0.
    k = data_z0[:, 0] * cosmo['h']
    pk = data_z0[:, -1] / (cosmo['h']**3)
    pk_ccl = ccl.halos.halomod_power_spectrum(cosmo, hmc, k, 1./(1+z), prf)
    tol = pk * HALOMOD_TOLERANCE
    err = np.abs(pk_ccl - pk)
    assert np.all(err <= tol)

    z = 1.
    k = data_z1[:, 0] * cosmo['h']
    pk = data_z1[:, -1] / (cosmo['h']**3)
    pk_ccl = ccl.halos.halomod_power_spectrum(cosmo, hmc, k, 1./(1+z), prf)
    tol = pk * HALOMOD_TOLERANCE
    err = np.abs(pk_ccl - pk)
    assert np.all(err <= tol)



__all__ = ("CosmicemuMTIVPk", "CosmicemuMTIIPk",)

import numpy as np
import os
import inspect
from abc import abstractmethod
from scipy.interpolate import interp1d

from .. import Pk2D
from . import EmulatorPk


class CosmicemuBase(EmulatorPk):
    """ Base class for non-linear power spectrum emulators
    from CosmicEmu.

    Args:
        kind (:obj:`str`): type of matter power spectrum to use.
            Options are `'tot'` (for the total matter power spectrum)
            or `'cb'` (for the CDM+baryons power spectrum).
    """
    def __init__(self, kind='tot'):
        # Parameter order
        self.pnames = ['omega_m', 'omega_b', 'sigma8', 'h', 'n_s',
                       'w_0', 'wtild', 'omega_nu']
        self.kind = kind

        # Data path
        data_path = os.path.join(os.path.dirname(
            os.path.abspath(inspect.stack()[0][1])), 'data')

        # Initialize emulator
        self._initialize(data_path)

    @abstractmethod
    def _initialize(self, data_path):
        """Initialize emulator from data folder
        """

    @abstractmethod
    def _get_pk_full(self, cosmo):
        """ Computes power spectrum at full grid of redshifts and k
        for this cosmology.
        """

    def _cosmo_to_x(self, cosmo):
        # Translates cosmology to an array of parameters used
        # by CosmicEmu.
        h = cosmo['h']
        omega_m = cosmo['Omega_m']*h**2
        omega_b = cosmo['Omega_b']*h**2
        sigma8 = cosmo['sigma8']
        if np.isnan(sigma8):
            raise ValueError("sigma8 must be provided to use CosmicEmu")
        wtild = (-cosmo['w0']-cosmo['wa'])**0.25
        omega_nu = cosmo.omega_x(1.0, 'neutrinos_massive')*h**2
        return np.array([omega_m, omega_b, sigma8, h,
                         cosmo['n_s'], cosmo['w0'], wtild, omega_nu])

    def _get_pk_at_a(self, cosmo, a):
        z, ks, pk = self._get_pk_full(cosmo)
        pki = interp1d(z, pk.T, kind='cubic')
        pks = pki(1./a-1).T
        return ks, pks

    def _get_pk2d(self, cosmo):
        z, ks, pk = self._get_pk_full(cosmo)
        return Pk2D(a_arr=1./(1+z), lk_arr=np.log(ks), pk_arr=np.log(pk),
                    is_logp=True, extrap_order_lok=1, extrap_order_hik=2)


class CosmicemuMTIIPk(CosmicemuBase):
    """ Nonlinear power spectrum emulator for CosmicEmu (Mira-Titan II,
    2017 version).

    This is an emulator of the non-linear matter power spectrum from
    N-body simulations, as a function of 8 cosmological parameters:
    (:math:`\\omega_m`, :math:`\\omega_b`, :math:`\\sigma_8`,
    :math:`n_s`, :math:`h`, :math:`w_0`, :math:`w_a`, and
    :math:`\\omega_\\nu`). The power spectrum is interpolated in this
    parameter space at a fixed grid of redshifts and wavenumbers.

    See `Lawrence et al. 2017 <https://arxiv.org/abs/1705.03388>`_.
    and https://github.com/lanl/CosmicEmu.

    Args:
        kind (:obj:`str`): type of matter power spectrum to use.
            Options are `'tot'` (for the total matter power spectrum)
            or `'cb'` (for the CDM+baryons power spectrum).
    """
    def __init__(self, kind='tot'):
        super().__init__(kind=kind)

    def _initialize(self, data_path):
        fname = os.path.join(data_path,
                             f"CosmicEmu_MTII_2017_P{self.kind}.npz")
        d = np.load(fname)
        self.x = d['x']
        self.xmin = d['xmin']
        self.xrange = d['xrange']
        self.xmax = d['xmax']
        self.K = d['K']
        self.w = [d['w1'], d['w2']]
        self.beta = [d['beta1'], d['beta2']]
        self.lamws = [d['lamws1'], d['lamws2']]
        self.lamz = [d['lamz1'], d['lamz2']]
        self.ks = d['mode']
        self.mean = d['mean']
        self.sd = d['sd']
        self.z = d['z'][::-1]
        self.nz = len(self.z)
        self.nk = len(self.ks)
        self.peta, self.nsims = [[w.shape[i] for w in self.w]
                                 for i in range(2)]

        # Compute KrigBasis
        x2diff = (self.x[:, None, :]-self.x[None, :, :])**2
        self.KrigBasis = []
        for j, (n, m) in enumerate(zip(self.peta, self.nsims)):
            x2d = x2diff[:m][:, :m, :]
            logc = np.sum(x2d*self.beta[j][:n, None, None, :],
                          axis=-1)
            sigma_sim = np.exp(-logc)/self.lamz[j][:, None, None]
            # Add to diagonal
            for i, lw in enumerate(self.lamws[j]):
                sigma_sim[i, :, :] += np.diag(np.full(m, 1./lw))
            kb = np.linalg.solve(sigma_sim, self.w[j])
            self.KrigBasis.append(kb)

    def _get_pk_full(self, cosmo):
        # Computes power spectrum at full grid of redshifts and k
        # for this cosmology.
        xstar = self._cosmo_to_x(cosmo)
        # Check for out of bounds
        out_of_bounds = (xstar < self.xmin) | (xstar > self.xmax)
        if np.any(out_of_bounds):
            for ip in np.where(out_of_bounds)[0]:
                pname = self.pnames[ip]
                pmin = self.xmin[ip]
                pmax = self.xmax[ip]
                raise ValueError(f'{pname} must be between {pmin} and {pmax}')
        # Standardize
        xstarstd = (xstar-self.xmin)/self.xrange

        # Compute covariance with new point
        wstar = []
        for i in range(2):
            xs = self.x[:self.nsims[i]]
            Sigmastar = np.exp(-np.sum(self.beta[i][:, None, :] *
                                       ((xs-xstarstd[None, :])**2)[None, :, :],
                                       axis=-1))/self.lamz[i][:, None]
            wstar.append(np.sum(Sigmastar*self.KrigBasis[i], axis=-1))
        wstar = np.concatenate(wstar)
        # Project and reshape
        ystaremu = (np.dot(self.K, wstar)*self.sd+self.mean).reshape([self.nz,
                                                                      self.nk])
        pk = 10**ystaremu*2*np.pi**2/self.ks**1.5
        return self.z, self.ks, pk


class CosmicemuMTIVPk(CosmicemuBase):
    """ Nonlinear power spectrum emulator for CosmicEmu (Mira-Titan IV,
    2022 version).

    This is an emulator of the non-linear matter power spectrum from
    N-body simulations, as a function of 8 cosmological parameters:
    (:math:`\\omega_m`, :math:`\\omega_b`, :math:`\\sigma_8`,
    :math:`n_s`, :math:`h`, :math:`w_0`, :math:`w_a`, and
    :math:`\\omega_\\nu`). The power spectrum is interpolated in this
    parameter space at a fixed grid of redshifts and wavenumbers.

    See `Moran et al. 2022 <https://arxiv.org/abs/2207.12345>`_.
    and https://github.com/lanl/CosmicEmu.

    Args:
        kind (:obj:`str`): type of matter power spectrum to use.
            Options are `'tot'` (for the total matter power spectrum)
            or `'cb'` (for the CDM+baryons power spectrum).
    """
    def __init__(self, kind='tot'):
        super().__init__(kind=kind)

    def _initialize(self, data_path):
        fname = os.path.join(data_path,
                             f"CosmicEmu_MTIV_2022_P{self.kind}.npz")
        d = np.load(fname)
        self.x = d['x']
        self.xmin = d['xmin']
        self.xrange = d['xrange']
        self.xmax = d['xmax']
        self.z = d['z']
        self.z_asc = d['z_asc']
        self.ks = d['mode']
        self.K = d['K']
        self.w = d['w']
        self.mean = d['mean']
        self.sd = d['sd']
        self.beta = d['beta']
        self.lamws = d['lamws']
        self.lamz = d['lamz']
        self.nsim, self.npar = self.x.shape
        self.nz = len(self.z)
        self.nk = len(self.ks)
        self.peta = len(self.lamz)

        # Compute KrigBasis
        x2diff = (self.x[:, None, :]-self.x[None, :, :])**2
        logc = np.sum(x2diff[None, :, :, :]*self.beta[:, None, None, :],
                      axis=-1)
        sigma_sim = np.exp(-logc)/self.lamz[:, None, None]
        # Add to diagonal
        for i, lw in enumerate(self.lamws):
            sigma_sim[i, :, :] += np.diag(np.full(self.nsim, 1./lw))
        self.KrigBasis = np.linalg.solve(sigma_sim, self.w)

        # Parameter order
        self.pnames = ['omega_m', 'omega_b', 'sigma8', 'h', 'n_s',
                       'w_0', 'wtild', 'omega_nu']

    def _get_pk_full(self, cosmo):
        # Computes power spectrum at full grid of redshifts and k
        # for this cosmology.
        xstar = self._cosmo_to_x(cosmo)
        # Check for out of bounds
        out_of_bounds = (xstar < self.xmin) | (xstar > self.xmax)
        if np.any(out_of_bounds):
            for ip in np.where(out_of_bounds)[0]:
                pname = self.pnames[ip]
                pmin = self.xmin[ip]
                pmax = self.xmax[ip]
                raise ValueError(f'{pname} must be between {pmin} and {pmax}')
        # Standardize
        xstarstd = (xstar-self.xmin)/self.xrange
        # Compute covariance with new point
        Sigmastar = np.exp(-np.sum(self.beta[:, None, :] *
                                   ((self.x-xstarstd[None, :])**2)[None, :, :],
                                   axis=-1))/self.lamz[:, None]
        # Project and reshape
        wstar = np.sum(Sigmastar*self.KrigBasis, axis=-1)
        ystaremu = (np.dot(self.K, wstar)*self.sd+self.mean).reshape([self.nz,
                                                                      self.nk])
        pk = 10**ystaremu*2*np.pi**2/self.ks**1.5
        return self.z, self.ks, pk



import numpy as np
import pytest
import pyccl as ccl


def get_pk_new(mf, c, cosmo, a, k, get_1h, get_2h):
    mdef = ccl.halos.MassDef('vir', 'matter')
    if mf == 'shethtormen':
        hmf = ccl.halos.MassFuncSheth99(mass_def=mdef,
                                        mass_def_strict=False,
                                        use_delta_c_fit=True)
        hbf = ccl.halos.HaloBiasSheth99(mass_def=mdef,
                                        mass_def_strict=False)
    elif mf == 'tinker10':
        hmf = ccl.halos.MassFuncTinker10(mass_def=mdef,
                                         mass_def_strict=False)
        hbf = ccl.halos.HaloBiasTinker10(mass_def=mdef,
                                         mass_def_strict=False)

    if c == 'constant_concentration':
        cc = ccl.halos.ConcentrationConstant(4., mass_def=mdef)
    elif c == 'duffy2008':
        cc = ccl.halos.ConcentrationDuffy08(mass_def=mdef)
    elif c == 'bhattacharya2011':
        cc = ccl.halos.ConcentrationBhattacharya13(mass_def=mdef)
    prf = ccl.halos.HaloProfileNFW(mass_def=mdef, concentration=cc)
    hmc = ccl.halos.HMCalculator(mass_function=hmf, halo_bias=hbf,
                                 mass_def=mdef)
    return ccl.halos.halomod_power_spectrum(cosmo, hmc, k, a, prf,
                                            get_1h=get_1h,
                                            get_2h=get_2h)


@pytest.mark.parametrize('mf_c', [['shethtormen', 'bhattacharya2011'],
                                  ['shethtormen', 'duffy2008'],
                                  ['shethtormen', 'constant_concentration'],
                                  ['tinker10', 'constant_concentration']])
def test_halomodel_choices_smoke(mf_c):
    mf, c = mf_c
    cosmo = ccl.Cosmology(
        Omega_c=0.27, Omega_b=0.045, h=0.67, sigma8=0.8, n_s=0.96,
        transfer_function='bbks', matter_power_spectrum='linear')
    a = 0.8
    k = np.geomspace(1E-2, 1, 10)
    p1h = get_pk_new(mf, c, cosmo, a, k, True, False)
    p2h = get_pk_new(mf, c, cosmo, a, k, False, True)
    pt = get_pk_new(mf, c, cosmo, a, k, True, True)

    assert np.all(np.isfinite(p1h))
    assert np.all(np.isfinite(p2h))
    assert np.all(np.isfinite(pt))
    assert np.allclose(pt, p1h+p2h, atol=0)



import numpy as np
import pytest
import pyccl as ccl
from .test_cclobject import check_eq_repr_hash


COSMO = ccl.Cosmology(
    Omega_c=0.27, Omega_b=0.045, h=0.67, sigma8=0.8, n_s=0.96,
    transfer_function='bbks', matter_power_spectrum='linear')
M200 = ccl.halos.MassDef200m
HMF = ccl.halos.MassFuncTinker10(mass_def=M200)
HBF = ccl.halos.HaloBiasTinker10(mass_def=M200)
CON = ccl.halos.ConcentrationDuffy08(mass_def=M200)
P1 = ccl.halos.HaloProfileNFW(mass_def=M200, concentration=CON,
                              fourier_analytic=True)
P2 = P1
P3 = ccl.halos.HaloProfilePressureGNFW(mass_def=M200)
PKC = ccl.halos.Profile2pt()
KK = np.geomspace(1E-3, 10, 32)
MM = np.geomspace(1E11, 1E15, 16)
AA = 1.0
PK2D = ccl.Pk2D.from_function(lambda k, a: a / k)


def test_prof2pt_smoke():
    uk_NFW = P1.fourier(COSMO, KK, MM, AA)
    uk_EIN = P2.fourier(COSMO, KK, MM, AA)
    # Variance
    cv_NN = PKC.fourier_2pt(COSMO, KK, MM, AA, P1)
    assert np.all(np.fabs((cv_NN - uk_NFW**2)) < 1E-10)

    # 2-point
    cv_NE = PKC.fourier_2pt(COSMO, KK, MM, AA, P1, prof2=P2)
    assert np.all(np.fabs((cv_NE - uk_NFW * uk_EIN)) < 1E-10)


def smoke_assert_pkhm_real(func):
    sizes = [(0, 0),
             (2, 0),
             (0, 2),
             (2, 3),
             (1, 3),
             (3, 1)]
    shapes = [(),
              (2,),
              (2,),
              (2, 3),
              (1, 3),
              (3, 1)]
    for (sa, sk), sh in zip(sizes, shapes):
        if sk == 0:
            k = 0.1
        else:
            k = np.logspace(-2., 0., sk)
        if sa == 0:
            a = 1.
        else:
            a = np.linspace(0.5, 1., sa)
        p = func(k, a)
        assert np.shape(p) == sh
        assert np.all(np.isfinite(p))


def test_HMIngredients_eq_repr_hash():
    # Test eq, repr, hash for the HMCalculator and its ingredients.
    # 1. Build a halo model calculator using the default parametrizations.
    HMC = ccl.halos.HMCalculator(
        mass_function="Tinker08", halo_bias="Tinker10", mass_def="200m")

    # 2. Define separate default halo model ingredients.
    MDEF = ccl.halos.MassDef200m
    HMF = ccl.halos.MassFuncTinker08(mass_def=MDEF)
    HBF = ccl.halos.HaloBiasTinker10(mass_def=MDEF)
    HMC2 = ccl.halos.HMCalculator(
        mass_function=HMF, halo_bias=HBF, mass_def=MDEF)  # equal
    HMC3 = ccl.halos.HMCalculator(
        mass_function="Press74", halo_bias="Sheth01", mass_def="fof")  # not eq

    # 3. Test equivalence.
    assert check_eq_repr_hash(MDEF, HMC.mass_def)
    assert check_eq_repr_hash(HMF, HMC.mass_function)
    assert check_eq_repr_hash(HBF, HMC.halo_bias)
    assert check_eq_repr_hash(HMC, HMC2)

    assert check_eq_repr_hash(MDEF, HMC3.mass_def, equal=False)
    assert check_eq_repr_hash(HMF, HMC3.mass_function, equal=False)
    assert check_eq_repr_hash(HBF, HMC3.halo_bias, equal=False)
    assert check_eq_repr_hash(HMC, HMC3, equal=False)


@pytest.mark.parametrize('norm', [True, False])
def test_pkhm_mean_profile_smoke(norm):
    hmc = ccl.halos.HMCalculator(mass_function=HMF, halo_bias=HBF,
                                 mass_def=M200, nM=2)

    def f(k, a):
        return ccl.halos.halomod_mean_profile_1pt(COSMO, hmc, k, a, P1)
    smoke_assert_pkhm_real(f)


@pytest.mark.parametrize('prof', [P1, P3])
def test_pkhm_bias_smoke(prof):
    hmc = ccl.halos.HMCalculator(mass_function=HMF, halo_bias=HBF,
                                 mass_def=M200, nM=2)

    def f(k, a):
        return ccl.halos.halomod_bias_1pt(COSMO, hmc, k, a, prof)
    smoke_assert_pkhm_real(f)


@pytest.mark.parametrize(
    "cv,pk,h1,h2,itg,p2",
    [(None, "linear", True, True, "simpson", None),
     (PKC, "linear", True, True, "simpson", None),
     (None, "linear", True, True, "simpson", P3),
     (None, "nonlinear", True, True, "simpson", None),
     (None, PK2D, True, True, "simpson", None),
     (None, None, True, True, "simpson", None),
     (None, "linear", False, True, "simpson", None),
     (None, "linear", True, False, "simpson", None),
     (None, "linear", False, False, "simpson", None),
     (None, "linear", True, True, "spline", None),
     (None, "linear", True, True, "simpson", P2)])
def test_pkhm_pk_smoke(cv, pk, h1, h2, itg, p2):
    hmc = ccl.halos.HMCalculator(mass_function=HMF, halo_bias=HBF,
                                 mass_def=M200, nM=2)

    def f(k, a):
        return ccl.halos.halomod_power_spectrum(COSMO, hmc, k, a,
                                                prof=P1, prof2=p2,
                                                prof_2pt=cv, p_of_k_a=pk,
                                                get_1h=h1, get_2h=h2)
    smoke_assert_pkhm_real(f)


def test_pkhm_pk2d():
    hmc = ccl.halos.HMCalculator(mass_function=HMF, halo_bias=HBF,
                                 mass_def=M200)
    k_arr = KK
    a_arr = np.linspace(0.3, 1, 10)
    pk_arr = ccl.halos.halomod_power_spectrum(COSMO, hmc, k_arr, a_arr, P1)

    # Input sampling
    pk2d = ccl.halos.halomod_Pk2D(COSMO, hmc, P1,
                                  lk_arr=np.log(k_arr), a_arr=a_arr)
    pk_arr_2 = pk2d(k_arr, a_arr, COSMO)
    assert np.all(np.fabs((pk_arr / pk_arr_2 - 1)).flatten()
                  < 1E-4)

    # Standard sampling
    pk2d = ccl.halos.halomod_Pk2D(COSMO, hmc, P1)
    pk_arr_2 = pk2d(k_arr, a_arr, COSMO)
    assert np.all(np.fabs((pk_arr / pk_arr_2 - 1)).flatten()
                  < 1E-4)

    # Testing profiles which are not equivalent (but very close)
    G1 = ccl.halos.HaloProfileHOD(mass_def=M200, concentration=CON,
                                  log10Mmin_0=12.00000)
    G2 = ccl.halos.HaloProfileHOD(mass_def=M200, concentration=CON,
                                  log10Mmin_0=11.99999)
    assert G1 != G2

    # I_1_1
    pk0 = ccl.halos.halomod_power_spectrum(COSMO, hmc, k_arr, a_arr, G1,
                                           prof2=G1)
    pk1 = ccl.halos.halomod_power_spectrum(COSMO, hmc, k_arr, a_arr, G1,
                                           prof2=G2)
    assert np.allclose(pk1, pk0, rtol=1e-4)

    # Profile normalization
    pk0 = ccl.halos.halomod_power_spectrum(COSMO, hmc, k_arr, a_arr, G1,
                                           prof2=G1)
    pk1 = ccl.halos.halomod_power_spectrum(COSMO, hmc, k_arr, a_arr, G1,
                                           prof2=G2)
    assert np.allclose(pk1, pk0, rtol=1e-4)

    # I_0_2 & I_1_2
    assert np.allclose(hmc.I_0_2(COSMO, KK, AA, P1, prof_2pt=PKC),
                       hmc.I_0_2(COSMO, KK, AA, P1, prof2=P1, prof_2pt=PKC),
                       rtol=0)
    assert np.allclose(hmc.I_1_2(COSMO, KK, AA, P1, prof_2pt=PKC),
                       hmc.I_1_2(COSMO, KK, AA, P1, prof2=P1, prof_2pt=PKC),
                       rtol=0)
    # I_0_22
    I0 = hmc.I_0_22(COSMO, KK, AA, P1, prof2=P1, prof3=P1, prof4=P1,
                    prof12_2pt=PKC, prof34_2pt=PKC)
    assert np.allclose(hmc.I_0_22(COSMO, KK, AA,
                                  P1, prof2=P1, prof3=P1, prof4=P1,
                                  prof12_2pt=PKC, prof34_2pt=None),
                       I0, rtol=0)
    assert np.allclose(hmc.I_0_22(COSMO, KK, AA,
                                  P1, prof2=P1, prof3=None, prof4=None,
                                  prof12_2pt=PKC, prof34_2pt=PKC),
                       I0, rtol=0)

    # 1h/2h transition
    def alpha0(a):  # no smoothing
        return 1.

    def alpha1(a):
        return 0.7

    pk0 = ccl.halos.halomod_power_spectrum(COSMO, hmc, k_arr, a_arr, P1,
                                           smooth_transition=None)
    pk1 = ccl.halos.halomod_power_spectrum(COSMO, hmc, k_arr, a_arr, P1,
                                           smooth_transition=alpha0)
    assert np.allclose(pk0, pk1, rtol=0)
    pk2 = ccl.halos.halomod_power_spectrum(COSMO, hmc, k_arr, a_arr, P1,
                                           smooth_transition=alpha1)
    assert np.all(pk2/pk0 > 1.)

    # 1-halo damping
    def ks0(a):  # no damping
        return 1e-16

    def ks1(a):  # fully suppressed
        return 1e16

    def ks2(a):  # reasonable
        return 0.04

    pk0 = ccl.halos.halomod_power_spectrum(COSMO, hmc, k_arr, a_arr, P1,
                                           suppress_1h=None, get_2h=False)
    pk1 = ccl.halos.halomod_power_spectrum(COSMO, hmc, k_arr, a_arr, P1,
                                           suppress_1h=ks0, get_2h=False)
    assert np.allclose(pk0, pk1, rtol=0)
    pk2 = ccl.halos.halomod_power_spectrum(COSMO, hmc, k_arr, a_arr, P1,
                                           suppress_1h=ks1, get_2h=False)
    assert np.allclose(pk2, 0, rtol=0)
    pk3 = ccl.halos.halomod_power_spectrum(COSMO, hmc, k_arr, a_arr, P1,
                                           suppress_1h=ks2, get_2h=False)
    fact = (k_arr/0.04)**4 / (1 + (k_arr/0.04)**4)
    assert np.allclose(pk3, pk0*fact, rtol=0)


def test_pkhm_errors():
    # Wrong integration
    with pytest.raises(ValueError):
        ccl.halos.HMCalculator(mass_function=HMF, halo_bias=HBF,
                               mass_def=M200, integration_method_M='Sampson')

    # Inconsistent mass definitions
    m200c = ccl.halos.MassDef.create_instance("200c")
    m200m = ccl.halos.MassDef.create_instance("200m")
    hmf = ccl.halos.MassFunc.create_instance("Tinker08", mass_def=m200c)
    hbf = ccl.halos.HaloBias.create_instance("Tinker10", mass_def=m200m)
    with pytest.raises(ValueError):
        ccl.halos.HMCalculator(mass_function=hmf, halo_bias=hbf,
                               mass_def=m200c)

    hmc = ccl.halos.HMCalculator(mass_function=HMF, halo_bias=HBF,
                                 mass_def=M200)

    # Wrong pk2d
    with pytest.raises(TypeError):
        ccl.halos.halomod_power_spectrum(COSMO, hmc, KK, AA, P1,
                                         p_of_k_a=KK)

    def func():
        pass

    # Wrong 1h/2h smoothing
    with pytest.raises(ValueError):
        ccl.halos.halomod_power_spectrum(COSMO, hmc, KK, AA, P1,
                                         smooth_transition=func, get_1h=False)

    # Wrong 1h damping
    with pytest.raises(ValueError):
        ccl.halos.halomod_power_spectrum(COSMO, hmc, KK, AA, P1,
                                         suppress_1h=func, get_1h=False)


def test_hmcalculator_from_string_smoke():
    hmc1 = ccl.halos.HMCalculator(
        mass_function=HMF, halo_bias=HBF, mass_def=M200)
    hmc2 = ccl.halos.HMCalculator(
        mass_function="Tinker10", halo_bias="Tinker10", mass_def="200m")
    for attr in ["mass_function", "halo_bias", "mass_def"]:
        assert getattr(hmc1, attr).name == getattr(hmc2, attr).name


def test_hmcalculator_from_string_raises():
    # Check that if the necessary arguments aren't provided, it raises.
    kw = {"mass_function": "Tinker10", "halo_bias": "Tinker10"}
    with pytest.raises(ValueError):
        # needs a mass_def to pass to the halo model ingredients
        ccl.halos.HMCalculator(**kw)
    # but this one is fine
    ccl.halos.HMCalculator(**kw, mass_def="200c")


def test_hmcalculator_inconsistent_mass_def_raises():
    # Check that HMCalculator complains for inconsistent mass definitions.
    hmc = ccl.halos.HMCalculator(
        mass_function="Tinker10", halo_bias="Tinker10", mass_def="200c")
    prof = ccl.halos.HaloProfilePressureGNFW(mass_def="500c")
    with pytest.raises(ValueError):
        hmc._check_mass_def(prof)



import numpy as np
import pyccl as ccl


def test_bocquet20_mf():
    d = np.load("./benchmarks/data/mf_bocquet20mf.py.npz")
    m_nu = ccl.nu_masses(Omega_nu_h2=float(d['Omnuh2']),
                         mass_split='equal')
    h = float(d['h'])
    cosmo = ccl.Cosmology(Omega_c=(float(d['Ommh2']) -
                                   float(d['Ombh2']) -
                                   float(d['Omnuh2']))/h**2,
                          Omega_b=float(d['Ombh2'])/h**2,
                          h=h, n_s=float(d['n_s']),
                          sigma8=float(d['sigma_8']),
                          m_nu=m_nu,
                          w0=float(d['w_0']),
                          wa=float(d['w_a']))
    mf = ccl.halos.MassFuncBocquet20()
    Ms = d['m']/h
    for z, mfp in zip(d['z'], d['mf']):
        mf_here = mf(cosmo, Ms, 1/(1+z))
        assert np.allclose(mfp*h**3*np.log(10), mf_here,
                           atol=0, rtol=1E-5)



import numpy as np
import pytest
import pyccl as ccl

EXTRAP_TYPES = [
    'none',
    'constant',
    'linx_liny',
    'linx_logy',
    'logx_liny',
    'logx_logy'
]
R_INI = 0.1
R_END = 100
R_ARR = np.linspace(R_INI, R_END, 512)


def test_resample_raises():
    with pytest.raises(ccl.CCLError):
        ccl.resample_array(R_ARR, R_ARR[:-1], R_ARR)
    with pytest.raises(ValueError):
        ccl.resample_array(R_ARR, R_ARR, R_ARR,
                           'linx_lagy', 'linx_logy')
    with pytest.raises(ValueError):
        ccl.resample_array(R_ARR, R_ARR, R_ARR,
                           'linx_logy', 'linx_lagy')


@pytest.mark.parametrize('extrap', EXTRAP_TYPES)
def test_resample_extrapolation(extrap):
    tilt = -1.
    offset = 1

    def f(r):
        if extrap == 'linx_liny':
            return r * tilt
        elif extrap == 'linx_logy':
            return np.exp(r * tilt)
        elif extrap == 'logx_liny':
            return tilt * np.log(r)
        elif extrap == 'logx_logy':
            return r**tilt
        else:
            trunc_ini = np.heaviside(r - R_INI, 0)
            trunc_end = 1 - np.heaviside(r - R_END, 0)
            trunc = trunc_ini * trunc_end
            return r * tilt * trunc + offset

    f_arr = f(R_ARR)
    r_ini_x = 0.01
    r_end_x = 200
    r_arr_x = np.geomspace(r_ini_x, r_end_x, 2048)
    if extrap == 'none':
        with pytest.raises(ccl.CCLError):
            ccl.resample_array(R_ARR, f_arr, r_arr_x,
                               extrap, extrap,
                               offset, offset)
    else:
        f_arr_x = ccl.resample_array(R_ARR, f_arr, r_arr_x,
                                     extrap, extrap,
                                     offset, offset)
        id_extrap = (r_arr_x > R_END) | (r_arr_x < R_INI)
        f_arr_x_pred = f(r_arr_x)
        res = np.fabs(f_arr_x / f_arr_x_pred - 1)
        assert np.all(res[id_extrap] < 1E-10)



import numpy as np
import pytest
import pyccl as ccl
from pyccl import UnlockInstance
from .test_cclobject import check_eq_repr_hash


COSMO = ccl.Cosmology(
    Omega_c=0.27, Omega_b=0.045, h=0.67, sigma8=0.8, n_s=0.96,
    transfer_function='bbks', matter_power_spectrum='linear')


def test_HaloProfile_eq_repr_hash():
    # Test eq, repr, hash for HaloProfile and Profile2pt.
    # 1. HaloProfile
    CM1 = ccl.halos.Concentration.from_name("Duffy08")()
    CM2 = ccl.halos.Concentration.from_name("Duffy08")()

    P1 = ccl.halos.HaloProfileHOD(mass_def=CM1.mass_def,
                                  concentration=CM1)
    P2 = ccl.halos.HaloProfileHOD(mass_def=CM2.mass_def,
                                  concentration=CM2)
    assert check_eq_repr_hash(CM1, CM2)
    assert check_eq_repr_hash(P1, P2)

    P1.update_parameters(log10Mmin_0=P1.log10Mmin_0/2)
    assert check_eq_repr_hash(P1, P2, equal=False)

    # 2. Profile2pt
    PCOV1 = ccl.halos.Profile2pt(r_corr=1.0)
    PCOV2 = ccl.halos.Profile2pt(r_corr=1.0)
    assert check_eq_repr_hash(PCOV1, PCOV2)

    PCOV2.update_parameters(r_corr=1.5)
    assert check_eq_repr_hash(PCOV1, PCOV2, equal=False)


def one_f(cosmo, M, a=1):
    return 1 if np.isscalar(M) else np.ones(M.size)


def smoke_assert_prof_real(profile, method='_real'):
    sizes = [(0, 0),
             (2, 0),
             (0, 2),
             (2, 3),
             (1, 3),
             (3, 1)]
    shapes = [(),
              (2,),
              (2,),
              (2, 3),
              (1, 3),
              (3, 1)]
    for (sm, sr), sh in zip(sizes, shapes):
        if sr == 0:
            r = 0.5
        else:
            # Don't include 0 to avoid 1/0 at origin.
            r = np.linspace(0.001, 1., sr)
        if sm == 0:
            m = 1E12
        else:
            m = np.geomspace(1E10, 1E14, sm)
        p = getattr(profile, method)(COSMO, r, m, 1)
        assert np.shape(p) == sh


def test_profiles_equal():
    cm = ccl.halos.ConcentrationDuffy08(mass_def='200m')
    p1 = ccl.halos.HaloProfileHOD(mass_def='200m',
                                  concentration=cm, log10Mmin_0=12.)

    # different profile types
    p2 = ccl.halos.HaloProfilePressureGNFW(mass_def='200m')
    assert p1 != p2

    # equal profiles
    p2 = p1
    assert p1 == p2

    # equivalent profiles
    cm2 = ccl.halos.ConcentrationDuffy08(mass_def='200m')
    p2 = ccl.halos.HaloProfileHOD(mass_def='200m',
                                  concentration=cm2, log10Mmin_0=12.)
    assert p1 == p2

    # different parameters
    p2 = ccl.halos.HaloProfileHOD(mass_def='200m', concentration=cm,
                                  log10Mmin_0=11.)
    assert p1 != p2

    # different mass-concentration
    cm2 = ccl.halos.ConcentrationConstant()
    p2 = ccl.halos.HaloProfileHOD(mass_def=cm2.mass_def, concentration=cm2,
                                  log10Mmin_0=12.)
    assert p1 != p2

    # different mass-concentration mass definition
    cm2 = ccl.halos.ConcentrationDuffy08(mass_def='200c')
    p2 = ccl.halos.HaloProfileHOD(mass_def='200c', concentration=cm2,
                                  log10Mmin_0=12.)
    assert p1 != p2

    # different FFTLog
    p2 = ccl.halos.HaloProfileHOD(mass_def='200m', concentration=cm,
                                  log10Mmin_0=12.)
    p2.update_precision_fftlog(**{"plaw_fourier": -2.0})
    assert p1 != p2

    # different FFTLog type
    assert p1.precision_fftlog != 1


@pytest.mark.parametrize('prof_class',
                         [ccl.halos.HaloProfileNFW,
                          ccl.halos.HaloProfileHernquist,
                          ccl.halos.HaloProfileEinasto])
def test_empirical_smoke(prof_class):
    c = ccl.halos.ConcentrationDuffy08(mass_def='200c')

    if prof_class in [ccl.halos.HaloProfileNFW,
                      ccl.halos.HaloProfileHernquist]:
        with pytest.raises(ValueError):
            p = prof_class(mass_def=c.mass_def, concentration=c,
                           projected_analytic=True,
                           truncated=True)
        with pytest.raises(ValueError):
            p = prof_class(mass_def=c.mass_def, concentration=c,
                           cumul2d_analytic=True,
                           truncated=True)
        p = prof_class(mass_def=c.mass_def, concentration=c)
        smoke_assert_prof_real(p, method='_fourier_analytic')
        smoke_assert_prof_real(p, method='_projected_analytic')
        smoke_assert_prof_real(p, method='_cumul2d_analytic')
    else:
        with pytest.raises(ValueError):
            p = prof_class(mass_def=c.mass_def, concentration=c,
                           projected_quad=True,
                           truncated=True)
        p = prof_class(mass_def=c.mass_def, concentration=c)
        smoke_assert_prof_real(p, method='_projected_quad')

    p = prof_class(mass_def=c.mass_def, concentration=c)
    smoke_assert_prof_real(p, method='real')
    smoke_assert_prof_real(p, method='projected')
    smoke_assert_prof_real(p, method='fourier')


def test_cib_smoke():
    c = ccl.halos.ConcentrationDuffy08(mass_def='200c')
    p = ccl.halos.HaloProfileCIBShang12(concentration=c, nu_GHz=217,
                                        mass_def='200c')
    beta_old = p.beta
    smoke_assert_prof_real(p, method='_real')
    smoke_assert_prof_real(p, method='_fourier')
    smoke_assert_prof_real(p, method='_fourier_variance')
    p.update_parameters(alpha=1.24, T0=20.0)
    assert p.alpha == 1.24
    assert p.T0 == 20.0
    assert p.beta == beta_old
    for n in ['alpha', 'T0', 'beta', 'gamma',
              's_z', 'Mmin', 'L0', 'siglog10M']:
        p.update_parameters(**{n: 1234.})
        assert getattr(p, n) == 1234.


def test_cib_2pt_raises():
    c = ccl.halos.ConcentrationDuffy08(mass_def='200c')
    p_cib = ccl.halos.HaloProfileCIBShang12(concentration=c, nu_GHz=217,
                                            mass_def='200c')
    p_tSZ = ccl.halos.HaloProfilePressureGNFW(mass_def='200c')
    p2pt = ccl.halos.Profile2ptCIB()
    with pytest.raises(TypeError):
        p2pt.fourier_2pt(COSMO, 0.1, 1E13, 1., p_tSZ)
    with pytest.raises(TypeError):
        p2pt.fourier_2pt(COSMO, 0.1, 1E13, 1., p_cib, prof2=p_tSZ)


def test_einasto_smoke():
    c = ccl.halos.ConcentrationDuffy08(mass_def='200c')
    p = ccl.halos.HaloProfileEinasto(mass_def='200c', concentration=c)
    for M in [1E14, [1E14, 1E15]]:
        alpha_from_cosmo = p._get_alpha(COSMO, M, 1)

        p.update_parameters(alpha=1.)
        alpha = p._get_alpha(COSMO, M, 1)
        assert np.ndim(M) == np.ndim(alpha)
        assert np.all(p._get_alpha(COSMO, M, 1) == np.full_like(M, 1.))

        p.update_parameters(alpha='cosmo')
        assert np.ndim(M) == np.ndim(alpha_from_cosmo)
        assert np.all(p._get_alpha(COSMO, M, 1) == alpha_from_cosmo)


def test_gnfw_smoke():
    p = ccl.halos.HaloProfilePressureGNFW(mass_def='200c')
    beta_old = p.beta
    smoke_assert_prof_real(p)
    p.update_parameters(mass_bias=0.7,
                        alpha=1.24)
    assert p.alpha == 1.24
    assert p.mass_bias == 0.7
    assert p.beta == beta_old
    for n in ['P0', 'P0_hexp', 'alpha',
              'beta', 'gamma', 'alpha_P',
              'c500', 'mass_bias', 'x_out']:
        p.update_parameters(**{n: 1.314159})
        assert getattr(p, n) == 1.314159


def test_gnfw_refourier():
    p = ccl.halos.HaloProfilePressureGNFW(mass_def='200c')
    # Create Fourier template
    p._integ_interp()
    p_f1 = p.fourier(COSMO, 1., 1E13, 1)
    # Check the Fourier profile gets recalculated
    p.update_parameters(alpha=1.32, c500=p.c500+0.1)
    p_f2 = p.fourier(COSMO, 1., 1E13, 1)
    assert p_f1 != p_f2


def test_hod_smoke():
    prof_class = ccl.halos.HaloProfileHOD
    c = ccl.halos.ConcentrationDuffy08(mass_def='200c')

    p = prof_class(mass_def=c.mass_def, concentration=c)
    smoke_assert_prof_real(p)
    smoke_assert_prof_real(p, method='_usat_real')
    smoke_assert_prof_real(p, method='_usat_fourier')
    smoke_assert_prof_real(p, method='_fourier')
    smoke_assert_prof_real(p, method='_fourier_variance')
    for n in ['log10Mmin_0', 'log10Mmin_p', 'log10M0_0', 'log10M0_p',
              'log10M1_0', 'log10M1_p', 'siglnM_0', 'siglnM_p',
              'fc_0', 'fc_p', 'alpha_0', 'alpha_p',
              'bg_0', 'bg_p', 'bmax_0', 'bmax_p',
              'a_pivot']:
        p.update_parameters(**{n: 1234.})
        assert getattr(p, n) == 1234.


@pytest.mark.parametrize('real_prof', [True, False])
def test_hod_ns_independent(real_prof):
    def func(prof):
        return prof._real if real_prof else prof._fourier

    c = ccl.halos.ConcentrationDuffy08(mass_def='200c')
    p1 = ccl.halos.HaloProfileHOD(mass_def='200c', concentration=c,
                                  log10Mmin_0=12.,
                                  ns_independent=False)
    p2 = ccl.halos.HaloProfileHOD(mass_def='200c', concentration=c,
                                  log10Mmin_0=12.,
                                  ns_independent=True)
    # M < Mmin
    f1 = func(p1)(COSMO, 0.01, 1e10, 1.)
    assert np.all(f1 == 0)
    f2 = func(p2)(COSMO, 0.01, 1e10, 1.)
    assert np.all(f2 > 0)
    # M > Mmin
    f1 = func(p1)(COSMO, 0.01, 1e14, 1.)
    f2 = func(p2)(COSMO, 0.01, 1e14, 1.)
    assert np.allclose(f1, f2, rtol=0)
    # M == Mmin
    f1 = func(p1)(COSMO, 0.01, 1e12, 1.)
    f2 = func(p2)(COSMO, 0.01, 1e12, 1.)
    assert np.allclose(2*f1, f2+0.5, rtol=0)

    if not real_prof:
        f1 = p1._fourier_variance(COSMO, 0.01, 1e10, 1)
        f2 = p2._fourier_variance(COSMO, 0.01, 1e10, 1)
        assert f2 > f1 == 0

    p1.update_parameters(ns_independent=True)
    assert p1.ns_independent is True


@pytest.mark.parametrize("ns_indep", [True, False])
def test_hod_normalization(ns_indep):
    # Test that the HOD normalization works as expected.
    cosmo = ccl.CosmologyVanillaLCDM(transfer_function="bbks")
    a_arr = np.linspace(0.5, 1.0, 8)
    hmc = ccl.halos.HMCalculator(
        mass_function="Tinker10", halo_bias="Tinker10", mass_def="200c")
    cm = ccl.halos.Concentration.create_instance(
        "Duffy08", mass_def=hmc.mass_def)
    prof = ccl.halos.HaloProfileHOD(mass_def='200c',
                                    concentration=cm, ns_independent=ns_indep)

    norm = np.array([prof.get_normalization(cosmo, a, hmc=hmc) for a in a_arr])

    def profile_norm(a):
        hmc._get_ingredients(cosmo, a, get_bf=False)
        uk0 = prof.fourier(cosmo, k=1e-5, M=hmc._mass, a=a).T
        return hmc._integrate_over_mf(uk0)

    norm_fourier = np.array([profile_norm(a) for a in a_arr])
    assert np.allclose(norm, norm_fourier, atol=0, rtol=1e-5)


def test_hod_2pt():
    pbad = ccl.halos.HaloProfilePressureGNFW(mass_def='200c')
    c = ccl.halos.ConcentrationDuffy08(mass_def='200c')
    pgood = ccl.halos.HaloProfileHOD(mass_def='200c', concentration=c)
    pgood_b = ccl.halos.HaloProfileHOD(mass_def='200c', concentration=c)
    p2 = ccl.halos.Profile2ptHOD()
    F0 = p2.fourier_2pt(COSMO, 1., 1e13, 1., pgood, prof2=pgood)
    assert np.allclose(p2.fourier_2pt(COSMO, 1., 1e13, 1., pgood, prof2=None),
                       F0, rtol=0)

    with pytest.raises(TypeError):
        p2.fourier_2pt(COSMO, 1., 1E13, 1., pbad)

    # doesn't raise because profiles are equivalent
    p2.fourier_2pt(COSMO, 1., 1E13, 1., pgood, prof2=pgood)

    p2.fourier_2pt(COSMO, 1., 1E13, 1., pgood, prof2=pgood_b)

    with pytest.raises(ValueError):
        pgood_b.update_parameters(log10M0_0=10.)
        p2.fourier_2pt(COSMO, 1., 1E13, 1., pgood, prof2=pgood_b)

    with pytest.raises(ValueError):
        p2.fourier_2pt(COSMO, 1., 1e13, 1., pgood, prof2=pbad)


def test_2pt_rcorr_smoke():
    c = ccl.halos.ConcentrationDuffy08(mass_def='200c')
    p = ccl.halos.HaloProfileNFW(mass_def='200c', concentration=c)
    F0 = ccl.halos.Profile2pt().fourier_2pt(COSMO, 1., 1e13, 1., p)
    p2pt = ccl.halos.Profile2pt(r_corr=0)
    F1 = p2pt.fourier_2pt(COSMO, 1., 1e13, 1., p)
    assert F0 == F1
    F2 = p2pt.fourier_2pt(COSMO, 1., 1e13, 1., p, prof2=p)
    assert F1 == F2

    p2pt.update_parameters(r_corr=-1.)
    assert p2pt.r_corr == -1.
    F3 = p2pt.fourier_2pt(COSMO, 1., 1e13, 1., p)
    assert F3 == 0


def get_nfw(real=False, fourier=False):
    # Return a subclass of the NFW profile with or without fourier analytic.
    NFW = type("NFW", (ccl.halos.HaloProfileNFW,), {})
    if real:
        NFW._real = ccl.halos.HaloProfileNFW._real
    if fourier:
        NFW._fourier = ccl.halos.HaloProfileNFW._fourier_analytic
    return NFW


@pytest.mark.parametrize("use_analytic", [True, False])
def test_nfw_accuracy(use_analytic):
    from scipy.special import sici

    tol = 1E-10 if use_analytic else 5E-3
    cM = ccl.halos.ConcentrationDuffy08(mass_def='200c')
    p = get_nfw(real=True, fourier=use_analytic)(mass_def='200c',
                                                 concentration=cM)
    M = 1E14
    a = 0.5
    c = cM(COSMO, M, a)
    r_Delta = cM.mass_def.get_radius(COSMO, M, a) / a
    r_s = r_Delta / c

    def fk(k):
        x = k * r_s
        Si1, Ci1 = sici((1 + c) * x)
        Si2, Ci2 = sici(x)
        P1 = 1 / (np.log(1+c) - c / (1 + c))
        P2 = np.sin(x) * (Si1 - Si2) + np.cos(x) * (Ci1 - Ci2)
        P3 = np.sin(c * x)/((1 + c) * x)
        return M * P1 * (P2 - P3)

    k_arr = np.logspace(-2, 2, 256) / r_Delta
    fk_arr = p.fourier(COSMO, k_arr, M, a)
    fk_arr_pred = fk(k_arr)
    # Normalize to  1 at low k
    res = np.fabs((fk_arr - fk_arr_pred) / M)
    assert np.all(res < tol)


def test_nfw_f2r():
    cM = ccl.halos.ConcentrationDuffy08(mass_def='200c')
    p1 = ccl.halos.HaloProfileNFW(mass_def='200c', concentration=cM)
    # We force p2 to compute the real-space profile
    # by FFT-ing the Fourier-space one.
    p2 = get_nfw(fourier=True)(mass_def='200c',
                               concentration=cM)
    p2.update_precision_fftlog(padding_hi_fftlog=1E3)

    M = 1E14
    a = 0.5
    r_Delta = cM.mass_def.get_radius(COSMO, M, a) / a

    r_arr = np.logspace(-2, 1, ) * r_Delta
    pr_1 = p1.real(COSMO, r_arr, M, a)
    pr_2 = p2.real(COSMO, r_arr, M, a)

    id_good = r_arr < r_Delta  # Profile is 0 otherwise
    res = np.fabs(pr_2[id_good] / pr_1[id_good] - 1)
    assert np.all(res < 0.01)


@pytest.mark.parametrize('fourier_analytic', [True, False])
def test_nfw_projected_accuracy(fourier_analytic):
    cM = ccl.halos.ConcentrationDuffy08(mass_def='200c')
    # Analytic projected profile
    p1 = ccl.halos.HaloProfileNFW(mass_def='200c', concentration=cM,
                                  truncated=False, projected_analytic=True)
    # FFTLog
    p2 = get_nfw(fourier=fourier_analytic)(mass_def='200c',
                                           concentration=cM, truncated=False)

    M = 1E14
    a = 0.5
    rt = np.logspace(-3, 2, 1024)
    srt1 = p1.projected(COSMO, rt, M, a)
    srt2 = p2.projected(COSMO, rt, M, a)

    res2 = np.fabs(srt2/srt1-1)
    assert np.all(res2 < 5E-3)


@pytest.mark.parametrize('fourier_analytic', [True, False])
def test_nfw_cumul2d_accuracy(fourier_analytic):
    cM = ccl.halos.ConcentrationDuffy08(mass_def='200c')
    # Analytic cumul2d profile
    p1 = ccl.halos.HaloProfileNFW(mass_def='200c', concentration=cM,
                                  truncated=False, cumul2d_analytic=True)
    # FFTLog
    p2 = get_nfw(fourier=fourier_analytic)(mass_def='200c',
                                           concentration=cM, truncated=False)

    M = 1E14
    a = 1.0
    rt = np.logspace(-3, 2, 1024)
    srt1 = p1.cumul2d(COSMO, rt, M, a)
    srt2 = p2.cumul2d(COSMO, rt, M, a)

    res2 = np.fabs(srt2/srt1-1)
    assert np.all(res2 < 5E-3)


def test_upd_fftlog_raises():
    # Verify that FFTLogParams is immutable unless changed in a control manner.
    prof = ccl.halos.HaloProfilePressureGNFW(mass_def='200c')
    new_params = {"hello_there": 0.}
    with pytest.raises(AttributeError):
        prof.update_precision_fftlog(**new_params)

    with pytest.raises(AttributeError):
        prof.precision_fftlog.plaw_projected = 1


@pytest.mark.parametrize("use_analytic", [True, False])
def test_hernquist_accuracy(use_analytic):
    from scipy.special import sici

    tol = 1E-10 if use_analytic else 5E-3
    cM = ccl.halos.ConcentrationDuffy08(mass_def='200c')
    p = ccl.halos.HaloProfileHernquist(mass_def='200c', concentration=cM,
                                       fourier_analytic=use_analytic)
    M = 1E14
    a = 0.5
    c = cM(COSMO, M, a)
    r_Delta = cM.mass_def.get_radius(COSMO, M, a) / a
    r_s = r_Delta / c

    def fk(k):
        x = k * r_s
        cp1 = c + 1
        Si1, Ci1 = sici(cp1 * x)
        Si2, Ci2 = sici(x)
        P1 = 1 / ((c / cp1)**2 / 2)
        P2 = x * np.sin(x) * (Ci1 - Ci2) - x * np.cos(x) * (Si1 - Si2)
        P3 = (-1 + np.sin(c * x) / (cp1**2 * x)
              + cp1 * np.cos(c * x) / (cp1**2))
        return M * P1 * (P2 - P3) / 2

    k_arr = np.logspace(-2, 2, 256) / r_Delta
    fk_arr = p.fourier(COSMO, k_arr, M, a)
    fk_arr_pred = fk(k_arr)
    # Normalize to  1 at low k
    res = np.fabs((fk_arr - fk_arr_pred) / M)
    assert np.all(res < tol)


def test_hernquist_f2r():
    cM = ccl.halos.ConcentrationDuffy08(mass_def='200c')
    p1 = ccl.halos.HaloProfileHernquist(mass_def='200c',
                                        concentration=cM)
    # We force p2 to compute the real-space profile
    # by FFT-ing the Fourier-space one.
    p2 = ccl.halos.HaloProfileHernquist(mass_def='200c',
                                        concentration=cM,
                                        fourier_analytic=True)
    with UnlockInstance(p2):
        p2._real = None
    p2.update_precision_fftlog(padding_hi_fftlog=1E3)

    M = 1E14
    a = 0.5
    r_Delta = cM.mass_def.get_radius(COSMO, M, a) / a

    r_arr = np.logspace(-2, 1, ) * r_Delta
    pr_1 = p1.real(COSMO, r_arr, M, a)
    pr_2 = p2.real(COSMO, r_arr, M, a)

    id_good = r_arr < r_Delta  # Profile is 0 otherwise
    res = np.fabs(pr_2[id_good] / pr_1[id_good] - 1)
    assert np.all(res < 0.01)


@pytest.mark.parametrize('fourier_analytic', [True, False])
def test_hernquist_projected_accuracy(fourier_analytic):
    cM = ccl.halos.ConcentrationDuffy08(mass_def='200c')
    # Analytic projected profile
    p1 = ccl.halos.HaloProfileHernquist(mass_def='200c',
                                        concentration=cM, truncated=False,
                                        projected_analytic=True)
    # FFTLog
    p2 = ccl.halos.HaloProfileHernquist(mass_def='200c',
                                        concentration=cM, truncated=False,
                                        fourier_analytic=fourier_analytic)

    M = 1E14
    a = 0.5
    rt = np.logspace(-3, 2, 1024)
    srt1 = p1.projected(COSMO, rt, M, a)
    srt2 = p2.projected(COSMO, rt, M, a)

    res2 = np.fabs(srt2/srt1-1)
    assert np.all(res2 < 5E-3)


@pytest.mark.parametrize('fourier_analytic', [True, False])
def test_hernquist_cumul2d_accuracy(fourier_analytic):
    cM = ccl.halos.ConcentrationDuffy08(mass_def='200c')
    # Analytic cumul2d profile
    p1 = ccl.halos.HaloProfileHernquist(mass_def='200c',
                                        concentration=cM, truncated=False,
                                        cumul2d_analytic=True)
    # FFTLog
    p2 = ccl.halos.HaloProfileHernquist(mass_def='200c',
                                        concentration=cM, truncated=False,
                                        fourier_analytic=fourier_analytic)
    M = 1E14
    a = 1.0
    rt = np.logspace(-3, 2, 1024)
    srt1 = p1.cumul2d(COSMO, rt, M, a)
    srt2 = p2.cumul2d(COSMO, rt, M, a)

    res2 = np.fabs(srt2/srt1-1)
    assert np.all(res2 < 5E-3)


def test_einasto_projected_accuracy():
    cM = ccl.halos.ConcentrationDuffy08(mass_def='200c')
    # projected profile from quad
    p1 = ccl.halos.HaloProfileEinasto(mass_def='200c',
                                      concentration=cM, truncated=False,
                                      projected_quad=True)
    # projected profile from FFTLog
    p2 = ccl.halos.HaloProfileEinasto(mass_def='200c',
                                      concentration=cM, truncated=False,
                                      projected_quad=False)
    # truncated projected profile from FFTLog
    p3 = ccl.halos.HaloProfileEinasto(mass_def='200c',
                                      concentration=cM, truncated=True,
                                      projected_quad=False)

    M = 1E14
    a = 0.5
    rt = np.logspace(-3, 2, 1024)
    srt1 = p1.projected(COSMO, rt, M, a)[:500]
    srt2 = p2.projected(COSMO, rt, M, a)[:500]
    srt3 = p3.projected(COSMO, rt, M, a)[:500]

    res1 = np.fabs(srt2/srt1-1)
    res2 = np.fabs(srt3/srt1-1)
    assert np.all(res1 < 6e-5)
    assert np.all(res2 < 6e-2)


def test_HaloProfile_abstractmethods():
    # Test that `HaloProfile` and its subclasses can't be instantiated if
    # either `_real` or `_fourier` have not been defined.
    with pytest.raises(TypeError):
        ccl.halos.HaloProfile()


def test_IA_update():
    cM = ccl.halos.ConcentrationDuffy08(mass_def="200m")
    p = ccl.halos.SatelliteShearHOD(concentration=cM, mass_def="200m")
    for attr in ['a1h', 'b', 'log10Mmin_0', 'log10Mmin_p',
                 'log10M0_0', 'log10M0_p', 'log10M1_0', 'log10M1_p',
                 'siglnM_0', 'siglnM_p', 'alpha_0', 'alpha_p',
                 'bg_0', 'bg_p', 'bmax_0', 'bmax_p', 'a_pivot',
                 'rmin', 'N_r', 'N_jn']:
        p.update_parameters(**{attr: -123})
        assert getattr(p, attr) == -123
    # Special cases
    p.update_parameters(lmax=5)
    assert p.lmax == 5
    p.update_parameters(ns_independent=True)
    assert p.ns_independent


def test_IA_profile():
    cosmo = ccl.Cosmology(Omega_c=0.27, Omega_b=0.045, h=0.67,
                          sigma8=0.83, n_s=0.96)
    k_arr = np.geomspace(1E-3, 1e3, 256)  # For evaluating
    cM = ccl.halos.ConcentrationDuffy08(mass_def="200m")

    # lmax too low
    with pytest.warns(ccl.CCLWarning):
        ccl.halos.SatelliteShearHOD(concentration=cM, lmax=1,
                                    mass_def="200m")

    # lmax too high
    with pytest.warns(ccl.CCLWarning):
        ccl.halos.SatelliteShearHOD(concentration=cM, lmax=14,
                                    mass_def="200m")

    # lmax odd
    assert (ccl.halos.SatelliteShearHOD(concentration=cM, mass_def="200m",
                                        lmax=7).lmax) % 2 == 0

    # Run with b!={0,2}
    assert (ccl.halos.SatelliteShearHOD(
        concentration=cM, b=-1.9, mass_def="200m",
        lmax=12)._angular_fl).shape == (6, 1)

    # Testing FFTLog accuracy vs simps and spline method.
    s_g_HOD1 = ccl.halos.SatelliteShearHOD(concentration=cM,
                                           mass_def="200m")
    s_g_HOD2 = ccl.halos.SatelliteShearHOD(concentration=cM,
                                           mass_def="200m",
                                           integration_method='simpson')
    s_g_HOD3 = ccl.halos.SatelliteShearHOD(concentration=cM,
                                           mass_def="200m",
                                           integration_method='spline')
    s_g1 = s_g_HOD1._usat_fourier(cosmo, k_arr, 1e13, 1.)
    s_g2 = s_g_HOD2._usat_fourier(cosmo, k_arr, 1e13, 1.)
    s_g3 = s_g_HOD3._usat_fourier(cosmo, k_arr, 1e13, 1.)
    assert np.all(np.abs((s_g1 - s_g2) / s_g2)) > 0.05
    assert np.all(np.abs((s_g3 - s_g2) / s_g3)) > 0.05

    # Wrong integration method
    with pytest.raises(ValueError):
        ccl.halos.SatelliteShearHOD(concentration=cM,
                                    mass_def="200m",
                                    integration_method="something_else")


def test_prefactor():
    cosmo = ccl.Cosmology(Omega_c=0.27, Omega_b=0.045, h=0.67,
                          sigma8=0.83, n_s=0.96)
    cM = ccl.halos.ConcentrationDuffy08(mass_def="200m")
    nM = ccl.halos.MassFuncTinker08(mass_def="200m")
    bM = ccl.halos.HaloBiasTinker10(mass_def="200m")
    hmc = ccl.halos.HMCalculator(mass_function=nM,
                                 halo_bias=bM, mass_def="200m")

    p = ccl.halos.HaloProfilePressureGNFW(mass_def="200m")  # a simple profile
    assert np.all(np.abs(p.get_normalization(cosmo, 1., hmc=hmc)-1.0 < 1e-10))

    p = ccl.halos.SatelliteShearHOD(concentration=cM, mass_def="200m")
    assert p.get_normalization(cosmo, 1., hmc=hmc) > 0



import numpy as np
import pyccl as ccl


def test_cibcl():
    # Read benchmarks
    bm = np.loadtxt("benchmarks/data/cib_class_sz_szpowerspectrum.txt",
                    unpack=True)

    # Initialize cosmology and halo model ingredients
    cosmo = ccl.Cosmology(Omega_b=0.05,
                          Omega_c=0.25,
                          h=0.7,
                          n_s=0.9645,
                          A_s=2.02E-9,
                          Neff=3.046, T_CMB=2.725)
    mdef = ccl.halos.MassDef200m
    cM = ccl.halos.ConcentrationDuffy08(mass_def=mdef)
    nM = ccl.halos.MassFuncTinker10(mass_def=mdef, norm_all_z=True)
    bM = ccl.halos.HaloBiasTinker10(mass_def=mdef)
    hmc = ccl.halos.HMCalculator(mass_function=nM, halo_bias=bM,
                                 mass_def=mdef)
    pr = ccl.halos.HaloProfileCIBShang12(concentration=cM, nu_GHz=217,
                                         Mmin=1E10, mass_def=mdef)
    pr.update_parameters(nu_GHz=217,
                         alpha=0.36,
                         T0=24.4,
                         beta=1.75,
                         gamma=1.7,
                         s_z=3.6,
                         log10Meff=12.6,
                         siglog10M=np.sqrt(0.5),
                         Mmin=1E10,
                         L0=6.4E-8)

    pr2pt = ccl.halos.Profile2ptCIB()

    # CIB tracer
    tr = ccl.CIBTracer(cosmo, z_min=0.07)

    # 3D power spectrum
    pk = ccl.halos.halomod_Pk2D(cosmo, hmc, pr, prof_2pt=pr2pt)

    # Angular power spectrum
    ls = bm[0]
    cl = ccl.angular_cl(cosmo, tr, tr, ell=ls, p_of_k_a=pk)
    dl = cl*ls*(ls+1)/(2*np.pi)

    # Compare
    assert np.all(np.fabs(dl/(bm[31]+bm[32])-1) < 1E-2)



import numpy as np
import pyccl as ccl

BEMULIN_TOLERANCE = 1e-3


def test_baccoemu_linear():
    bemu = ccl.BaccoemuLinear()
    cosmo = ccl.Cosmology(
        Omega_c=0.27,
        Omega_b=0.05,
        h=0.67,
        sigma8=0.83,
        n_s=0.96,
        Neff=3.046,
        mass_split='normal',
        m_nu=0.1,
        Omega_g=0,
        Omega_k=0,
        w0=-1,
        wa=0,
        transfer_function=bemu)

    data = np.loadtxt("./benchmarks/data/baccoemu_linear.txt")

    k = data[:, 0] * cosmo['h']
    pk = data[:, 1] / cosmo['h']**3
    a = 1

    linpk = cosmo.get_linear_power()
    err = np.abs(pk / linpk(k, a) - 1)
    assert np.allclose(err, 0, atol=BEMULIN_TOLERANCE, rtol=0)

    ktest, pktest = bemu.get_pk_at_a(cosmo, 1.0)
    pktest = np.exp(np.interp(np.log(k), np.log(ktest), np.log(pktest)))
    err = np.abs(pktest / pk - 1)
    assert np.allclose(err, 0, atol=BEMULIN_TOLERANCE, rtol=0)


def test_baccoemu_linear_A_s():
    bemu = ccl.BaccoemuLinear()
    cosmo = ccl.Cosmology(
        Omega_c=0.27,
        Omega_b=0.05,
        h=0.67,
        A_s=2.2194e-9,
        n_s=0.96,
        Neff=3.046,
        mass_split='normal',
        m_nu=0.1,
        Omega_g=0,
        Omega_k=0,
        w0=-1,
        wa=0,
        transfer_function=bemu)

    data = np.loadtxt("./benchmarks/data/baccoemu_linear.txt")

    k = data[:, 0] * cosmo['h']
    pk = data[:, 1] / cosmo['h']**3
    a = 1

    linpk = cosmo.get_linear_power()
    err = np.abs(pk / linpk(k, a) - 1)
    assert np.allclose(err, 0, atol=BEMULIN_TOLERANCE, rtol=0)

    ktest, pktest = bemu.get_pk_at_a(cosmo, 1.0)
    pktest = np.exp(np.interp(np.log(k), np.log(ktest), np.log(pktest)))
    err = np.abs(pktest / pk - 1)
    assert np.allclose(err, 0, atol=BEMULIN_TOLERANCE, rtol=0)



__all__ = ("NeutrinoMassSplits", "nu_masses",)

from enum import Enum
from numbers import Real
from typing import Iterable

import numpy as np
from scipy.optimize import root

from . import physical_constants as const


class NeutrinoMassSplits(Enum):
    """Enumeration listing all allowed neutrino mass
    split types.

    - 'sum': sum of masses.
    - 'single': single massive neutrino.
    - 'equal': total mass distributed equally among 3 species.
    - 'normal': normal hierarchy.
    - 'inverted': inverted hierarchy.
    - 'list': a list of 3 different masses is passed.
    """
    SUM = 'sum'
    SINGLE = 'single'
    EQUAL = 'equal'
    NORMAL = 'normal'
    INVERTED = 'inverted'
    LIST = 'list'  # placeholder for backwards-compatibility


def nu_masses(*, Omega_nu_h2=None, mass_split, m_nu=None):
    """Returns the neutrinos mass(es) for a given value of
    :math:`\\Omega_\\nu h^2`, according to the splitting convention
    specified by the user.

    Args:
        Omega_nu_h2 (:obj:`float`): Neutrino energy density at z=0 times
            :math:`h^2`.
        mass_split (:obj:`str`): indicates how the masses should be split up
            Should be one of 'normal', 'inverted', 'equal' or 'sum'.
        m_nu (:obj:`float` or array_like):
            Mass in eV of the massive neutrinos present.
            If a sequence is passed, it is assumed that the elements of the
            sequence represent the individual neutrino masses.

    Returns:
        :obj:`float` or `array`: Neutrino mass(es) corresponding to this
        :math:`\\Omega_\\nu h^2`.
    """
    if m_nu is None:
        m_nu = 93.14 * Omega_nu_h2
    return _get_neutrino_masses(m_nu=m_nu, mass_split=mass_split)


def _get_neutrino_masses(*, m_nu, mass_split):
    """
    """
    if isinstance(m_nu, Real) and m_nu == 0:  # no massive neutrinos
        return np.array([])
    if isinstance(m_nu, Iterable):  # input was list
        return np.asarray(m_nu).copy()

    split = NeutrinoMassSplits

    if split(mass_split) == split.SUM:
        return m_nu
    if split(mass_split) == split.SINGLE:
        return np.atleast_1d(m_nu)
    if split(mass_split) == split.EQUAL:
        return np.full(3, m_nu/3)

    c = const
    D12, D13p, D13n = c.DELTAM12_sq, c.DELTAM13_sq_pos, c.DELTAM13_sq_neg

    def M_nu(m, D13):
        m2 = m * m
        return np.array([m.sum()-m_nu, m2[1]-m2[0]-D12, m2[2]-m2[0]-D13])

    def check_mnu(val):
        if m_nu < val:
            raise ValueError(f"m_nu < {val} incompatible with mass hierarchy")

    if split(mass_split) == split.NORMAL:
        check_mnu(np.sqrt(D12) + np.sqrt(D13p))
        x0 = [0, np.sqrt(D12), np.sqrt(D13p)]
        return root(M_nu, x0, args=(D13p,)).x
    if split(mass_split) == split.INVERTED:
        check_mnu(np.sqrt(-(D13n + D12)) + np.sqrt(-D13n))
        x0 = [0, np.sqrt(-(D13n + D12)), np.sqrt(-D13n)]
        return root(M_nu, x0, args=(D13n,)).x



import numpy as np
import pyccl as ccl
import time
import os
import pytest


@pytest.fixture(scope='module', params=['analytic', 'histo'])
def set_up(request):
    t0 = time.time()

    nztyp = request.param
    dirdat = os.path.dirname(__file__) + '/data/'
    ccl.gsl_params.LENSING_KERNEL_SPLINE_INTEGRATION = False
    ccl.gsl_params.INTEGRATION_LIMBER_EPSREL = 1E-4
    ccl.gsl_params.INTEGRATION_EPSREL = 1E-4
    cosmo = ccl.Cosmology(Omega_c=0.30, Omega_b=0.00, Omega_g=0, Omega_k=0,
                          h=0.7, sigma8=0.8, n_s=0.96, Neff=0, m_nu=0.0,
                          w0=-1, wa=0, T_CMB=2.7, transfer_function='bbks',
                          matter_power_spectrum='linear')

    # ell-arrays
    nls = 541
    ells = np.zeros(nls)
    ells[:50] = np.arange(50)+2
    ells[50:] = ells[49] + 6 * (np.arange(nls-50) + 1)
    fl_one = np.ones(nls)
    fl_dl = (ells + 0.5)**2/np.sqrt((ells + 2.) * (ells + 1.) *
                                    ells * (ells - 1.))
    fl_ll = fl_dl**2
    fl_lc = ells * (ells + 1)/np.sqrt((ells + 2.) * (ells + 1.) *
                                      ells * (ells - 1.))
    fl_li = 2 * fl_dl
    lfacs = {'ells': ells,
             'fl_one': fl_one,
             'fl_dl': fl_dl,
             'fl_ll': fl_ll,
             'fl_lc': fl_lc,
             'fl_li': fl_li}

    # Initialize tracers
    if nztyp == 'analytic':
        # Analytic case
        zmean_1 = 1.0
        sigz_1 = 0.15
        zmean_2 = 1.5
        sigz_2 = 0.15
        z1, tmp_a1 = np.loadtxt(dirdat + "ia_amp_analytic_1.txt", unpack=True)
        z2, tmp_a2 = np.loadtxt(dirdat + "ia_amp_analytic_2.txt", unpack=True)
        pz1 = np.exp(-0.5 * ((z1 - zmean_1) / sigz_1)**2)
        pz2 = np.exp(-0.5 * ((z2 - zmean_2) / sigz_2)**2)
    elif nztyp == 'histo':
        # Histogram case
        z1, pz1 = np.loadtxt(dirdat + "bin1_histo.txt", unpack=True)[:, 1:]
        _, tmp_a1 = np.loadtxt(dirdat + "ia_amp_histo_1.txt", unpack=True)
        z2, pz2 = np.loadtxt(dirdat + "bin2_histo.txt", unpack=True)[:, 1:]
        _, tmp_a2 = np.loadtxt(dirdat + "ia_amp_histo_2.txt", unpack=True)
    else:
        raise ValueError("Wrong Nz type " + nztyp)
    bz = np.ones_like(pz1)

    # Renormalize the IA amplitude to be consistent with A_IA
    D1 = ccl.growth_factor(cosmo, 1./(1+z1))
    D2 = ccl.growth_factor(cosmo, 1./(1+z2))
    rho_m = ccl.physical_constants.RHO_CRITICAL * cosmo['Omega_m']
    a1 = - tmp_a1 * D1 / (5e-14 * rho_m)
    a2 = - tmp_a2 * D2 / (5e-14 * rho_m)

    # Initialize tracers
    trc = {}
    trc['g1'] = ccl.NumberCountsTracer(cosmo, has_rsd=False,
                                       dndz=(z1, pz1),
                                       bias=(z2, bz))
    trc['g2'] = ccl.NumberCountsTracer(cosmo, has_rsd=False,
                                       dndz=(z2, pz2),
                                       bias=(z2, bz))
    trc['l1'] = ccl.WeakLensingTracer(cosmo, dndz=(z1, pz1))
    trc['l2'] = ccl.WeakLensingTracer(cosmo, dndz=(z2, pz2))
    trc['i1'] = ccl.WeakLensingTracer(cosmo, dndz=(z1, pz1),
                                      has_shear=False,
                                      ia_bias=(z1, a1))
    trc['i2'] = ccl.WeakLensingTracer(cosmo, dndz=(z2, pz2),
                                      has_shear=False,
                                      ia_bias=(z2, a2))
    trc['ct'] = ccl.CMBLensingTracer(cosmo, z_source=1100.)

    # Read benchmarks
    def read_bm(fname):
        _, cl = np.loadtxt(fname, unpack=True)
        return cl[ells.astype('int')]

    pre = dirdat + 'run_'
    post = nztyp + "_log_cl_"
    bms = {}
    bms['dd_11'] = read_bm(pre + 'b1b1' + post + 'dd.txt')
    bms['dd_12'] = read_bm(pre + 'b1b2' + post + 'dd.txt')
    bms['dd_22'] = read_bm(pre + 'b2b2' + post + 'dd.txt')
    bms['dl_11'] = read_bm(pre + 'b1b1' + post + 'dl.txt')
    bms['dl_12'] = read_bm(pre + 'b1b2' + post + 'dl.txt')
    bms['dl_21'] = read_bm(pre + 'b2b1' + post + 'dl.txt')
    bms['dl_22'] = read_bm(pre + 'b2b2' + post + 'dl.txt')
    bms['di_11'] = read_bm(pre + 'b1b1' + post + 'di.txt')
    bms['di_12'] = read_bm(pre + 'b1b2' + post + 'di.txt')
    bms['di_21'] = read_bm(pre + 'b2b1' + post + 'di.txt')
    bms['di_22'] = read_bm(pre + 'b2b2' + post + 'di.txt')
    bms['dc_1'] = read_bm(pre + 'b1b1' + post + 'dc.txt')
    bms['dc_2'] = read_bm(pre + 'b2b2' + post + 'dc.txt')
    bms['ll_11'] = read_bm(pre + 'b1b1' + post + 'll.txt')
    bms['ll_12'] = read_bm(pre + 'b1b2' + post + 'll.txt')
    bms['ll_22'] = read_bm(pre + 'b2b2' + post + 'll.txt')
    bms['li_11'] = read_bm(pre + 'b1b1' + post + 'li.txt')
    bms['li_22'] = read_bm(pre + 'b2b2' + post + 'li.txt')
    bms['lc_1'] = read_bm(pre + 'b1b1' + post + 'lc.txt')
    bms['lc_2'] = read_bm(pre + 'b2b2' + post + 'lc.txt')
    bms['ii_11'] = read_bm(pre + 'b1b1' + post + 'ii.txt')
    bms['ii_12'] = read_bm(pre + 'b1b2' + post + 'ii.txt')
    bms['ii_22'] = read_bm(pre + 'b2b2' + post + 'ii.txt')
    bms['cc'] = read_bm(pre + 'log_cl_cc.txt')
    print('init and i/o time:', time.time() - t0)

    ccl.gsl_params.reload()

    return cosmo, trc, lfacs, bms


@pytest.mark.parametrize("t1,t2,bm,a1b1,a1b2,a2b1,a2b2,fl",
                         [('g1', 'g1', 'dd_11',   # NC1-NC1
                           'dd_11', 'dd_11', 'dd_11', 'dd_11',
                           'fl_one'),
                          ('g1', 'g2', 'dd_12',   # NC1-NC2
                           'dd_11', 'dd_12', 'dd_12', 'dd_22',
                           'fl_one'),
                          ('g2', 'g2', 'dd_22',   # NC2-NC2
                           'dd_22', 'dd_22', 'dd_22', 'dd_22',
                           'fl_one'),
                          ('g1', 'l1', 'dl_11',   # NC1-WL1
                           'dd_11', 'dl_11', 'dl_11', 'll_11',
                           'fl_dl'),
                          ('g1', 'l2', 'dl_12',   # NC1-WL2
                           'dd_11', 'dl_12', 'dl_12', 'll_22',
                           'fl_dl'),
                          ('g2', 'l1', 'dl_21',   # NC2-WL1
                           'dd_22', 'dl_21', 'dl_21', 'll_22',
                           'fl_dl'),
                          ('g2', 'l2', 'dl_22',   # NC2-WL2
                           'dd_22', 'dl_22', 'dl_22', 'ii_22',
                           'fl_dl'),
                          ('g1', 'i1', 'di_11',   # NC1-IA1
                           'dd_11', 'di_11', 'di_11', 'ii_11',
                           'fl_dl'),
                          ('g1', 'i2', 'di_12',   # NC1-IA2
                           'dd_11', 'di_12', 'di_12', 'ii_22',
                           'fl_dl'),
                          ('g2', 'i1', 'di_21',   # NC2-IA1
                           'dd_22', 'di_21', 'di_21', 'ii_22',
                           'fl_dl'),
                          ('g2', 'i2', 'di_22',   # NC2-IA2
                           'dd_22', 'di_22', 'di_22', 'ii_22',
                           'fl_dl'),
                          ('g1', 'ct', 'dc_1',   # NC1-CMBL
                           'dd_11', 'dc_1', 'dc_1', 'cc',
                           'fl_one'),
                          ('g2', 'ct', 'dc_2',   # NC2-CMBL
                           'dd_22', 'dc_2', 'dc_2', 'cc',
                           'fl_one'),
                          ('l1', 'l1', 'll_11',   # WL1-WL1
                           'll_11', 'll_11', 'll_11', 'll_11',
                           'fl_ll'),
                          ('l1', 'l2', 'll_12',   # WL1-WL2
                           'll_11', 'll_12', 'll_12', 'll_22',
                           'fl_ll'),
                          ('l2', 'l2', 'll_22',   # WL2-WL2
                           'll_22', 'll_22', 'll_22', 'll_22',
                           'fl_ll'),
                          ('l1', 'i1', 'li_11',   # WL1-IA1
                           'ii_11', 'li_11', 'li_11', 'll_11',
                           'fl_li'),
                          ('l2', 'i2', 'li_22',   # WL2-IA2
                           'ii_22', 'li_22', 'li_22', 'll_22',
                           'fl_li'),
                          ('l1', 'ct', 'lc_1',   # WL1-CMBL
                           'll_11', 'lc_1', 'lc_1', 'cc',
                           'fl_lc'),
                          ('l2', 'ct', 'lc_2',   # WL2-CMBL
                           'll_22', 'lc_2', 'lc_2', 'cc',
                           'fl_lc'),
                          ('i1', 'i1', 'ii_11',   # IA1-IA1
                           'll_11', 'll_11', 'll_11', 'll_11',
                           'fl_ll'),
                          ('i1', 'i2', 'ii_12',   # IA1-IA2
                           'll_11', 'll_12', 'll_12', 'll_22',
                           'fl_ll'),
                          ('i2', 'i2', 'ii_22',   # IA2-IA2
                           'll_22', 'll_22', 'll_22', 'll_22',
                           'fl_ll')])
def test_cells(set_up, t1, t2, bm,
               a1b1, a1b2, a2b1, a2b2, fl):
    cosmo, trcs, lfc, bmk = set_up
    cl = ccl.angular_cl(cosmo, trcs[t1], trcs[t2], lfc['ells'],
                        limber_integration_method='qag_quad') * lfc[fl]
    el = np.sqrt((bmk[a1b1] * bmk[a2b2] + bmk[a1b2] * bmk[a2b1]) /
                 (2 * lfc['ells'] + 1.))
    assert np.all(np.fabs(cl - bmk[bm]) < 0.1 * el)


@pytest.mark.parametrize("t1,t2,bm,a1b1,a1b2,a2b1,a2b2,fl",
                         [('g1', 'g1', 'dd_11',   # NC1-NC1
                           'dd_11', 'dd_11', 'dd_11', 'dd_11',
                           'fl_one'),
                          ('g1', 'g2', 'dd_12',   # NC1-NC2
                           'dd_11', 'dd_12', 'dd_12', 'dd_22',
                           'fl_one'),
                          ('l2', 'l2', 'll_22',   # WL2-WL2
                           'll_22', 'll_22', 'll_22', 'll_22',
                           'fl_ll')])
def test_cells_spline(set_up, t1, t2, bm,
                      a1b1, a1b2, a2b1, a2b2, fl):
    cosmo, trcs, lfc, bmk = set_up
    cl = ccl.angular_cl(cosmo, trcs[t1], trcs[t2], lfc['ells'],
                        limber_integration_method='spline') * lfc[fl]
    el = np.sqrt((bmk[a1b1] * bmk[a2b2] + bmk[a1b2] * bmk[a2b1]) /
                 (2 * lfc['ells'] + 1.))
    assert np.all(np.fabs(cl - bmk[bm]) < 0.2 * el)
